/*
 *  CageGL.java
 *  Copyright Simon Hern 2009
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package Space;

import java.nio.*;
import javax.media.opengl.*;

// graphical representation of the Cage around a planetary system
public class CageGL {

  // line details
  private static final float kLineColour[] = { 0.5f, 0.5f, 0.5f, 1.0f },
                             kFogColour[]  = { 0.01f, 0.01f, 0.01f, 1.0f };
  
  // lines are only visible close up
  private static final float kFogStart = 0.1f,
                             kFogEnd   = 0.5f;
  
  // amount the triangular pattern pulls back from the icosahedron edges
  private static final float kPatternScale = 0.99f;
  
  // number of divisions in the triangular pattern
  private static final int kNumDivisions = 15;
    
  // (x,y,z) coordinates of vertices in one face of the triangular pattern
  private static FloatBuffer kVertexArray;
  
  // arrays of edges in one face of the triangular pattern
  private static ShortBuffer kEdgeArray;

  // 0 => direct, 1 => arrays, 2 => buffer objects
  private static int kDisplayMethod = 1;
  
  // display list (for unit icosahedron)
  private static int kDisplayList = 0;

  // prepare resources (display list, etc.)
  static public void init(GL2 gl) {

    if ( kDisplayList != 0 ) return; // already initialized

    if ( !Env.useBufferObjs() && kDisplayMethod == 2 ) kDisplayMethod = 1;
    
    buildArrays();
    buildDisplayList(gl);

  } // init()
  
  // discard any resources
  static public void dispose(GL2 gl) {
  } // dispose()
  
  // draw the object (unit radius, centred on origin)
  static public void display(GL2 gl) {
    
    assert( kDisplayList > 0 );
    gl.glCallList(kDisplayList);
    
  } // display()

  // define the vertex and edge arrays for the triangular pattern
  static private void buildArrays() {

    final float v0[] = Icosahedron.vertex(0),
                v1[] = Icosahedron.vertex(1),
                v2[] = Icosahedron.vertex(8);
    
    final float x0 = v0[0] + (1.0f-kPatternScale)*(v1[0] + v2[0] - 2*v0[0])/3,
                y0 = v0[1] + (1.0f-kPatternScale)*(v1[1] + v2[1] - 2*v0[1])/3,
                z0 = v0[2] + (1.0f-kPatternScale)*(v1[2] + v2[2] - 2*v0[2])/3;
    final float dx1 = kPatternScale*( v1[0] - v0[0] ),
                dy1 = kPatternScale*( v1[1] - v0[1] ),
                dz1 = kPatternScale*( v1[2] - v0[2] );
    final float dx2 = kPatternScale*( v2[0] - v0[0] ),
                dy2 = kPatternScale*( v2[1] - v0[1] ),
                dz2 = kPatternScale*( v2[2] - v0[2] );
        
    final int numVerts = (kNumDivisions*(kNumDivisions+1))/2;
    kVertexArray = Env.allocateFloatBuffer(3*numVerts);
    
    for ( int j = 0 ; j < kNumDivisions ; j++ ) {
      for ( int i = 0 ; i < kNumDivisions-j ; i++ ) {
        final float s1 = i/(kNumDivisions-1.0f),
                    s2 = j/(kNumDivisions-1.0f);
        kVertexArray.put(x0 + s1*dx1 + s2*dx2);
        kVertexArray.put(y0 + s1*dy1 + s2*dy2);
        kVertexArray.put(z0 + s1*dz1 + s2*dz2);
      }
    }
    
    final int numEdges = (3*kNumDivisions*(kNumDivisions-1))/2;
    kEdgeArray = Env.allocateShortBuffer(2*numEdges);
    
    for ( int j = 0 ; j < kNumDivisions-1 ; j++ ) {
      for ( int i = 0 ; i < kNumDivisions-1-j ; i++ ) {
        kEdgeArray.put( (short)vertexIndex(i,j) );
        kEdgeArray.put( (short)vertexIndex(i+1,j) );

        kEdgeArray.put( (short)vertexIndex(j,i) );
        kEdgeArray.put( (short)vertexIndex(j,i+1) );

        kEdgeArray.put( (short)vertexIndex(kNumDivisions-1-j-i,i) );
        kEdgeArray.put( (short)vertexIndex(kNumDivisions-2-j-i,i+1) );
      }
    }
    
  } // buildArrays()
  
  // create a display list for the whole cage
  static private void buildDisplayList(GL2 gl) {
    
    kDisplayList = gl.glGenLists(1);
    gl.glNewList(kDisplayList, GL2.GL_COMPILE);
    gl.glPushMatrix();
    
    gl.glDisable(GL2.GL_LIGHTING);
    gl.glEnable(GL.GL_BLEND);
    gl.glEnable(GL2.GL_FOG);
    
    gl.glEnable(GL.GL_LINE_SMOOTH);
    gl.glColor4fv(kLineColour, 0);

    gl.glFogi(GL2.GL_FOG_MODE, GL.GL_LINEAR);
    gl.glFogfv(GL2.GL_FOG_COLOR, kFogColour, 0);
    gl.glFogf(GL2.GL_FOG_START, kFogStart*Env.worldRadius());
    gl.glFogf(GL2.GL_FOG_END, kFogEnd*Env.worldRadius());

    bindArrays(gl);
    drawPattern(gl);
    unbindArrays(gl);
    
    gl.glDisable(GL2.GL_FOG);
    gl.glDisable(GL.GL_BLEND);
    gl.glEnable(GL2.GL_LIGHTING);
    
    gl.glPopMatrix();
    gl.glEndList();

    assert( kDisplayList > 0 );
  
  } // buildDisplayList()
  
  // prepare arrays or buffers for drawing
  static private void bindArrays(GL2 gl) {
    
    if ( kDisplayMethod == 0 ) return;

    assert( kDisplayMethod != 2 ); // not implemented yet
    
    kVertexArray.rewind();
    gl.glEnableClientState(GL2.GL_VERTEX_ARRAY);
    gl.glVertexPointer(3, GL.GL_FLOAT, 0, kVertexArray);
    
  } // beforeDisplay()
  
  // tidy up after drawing
  static private void unbindArrays(GL2 gl) {
    
    if ( kDisplayMethod == 0 ) return;
    
    assert( kDisplayMethod != 2 ); // not implemented yet
    
    gl.glDisableClientState(GL2.GL_VERTEX_ARRAY);
    
  } // afterDisplay()
  
  // draw the triangular pattern on the icosahedron (unit radius)
  static private void drawPattern(GL2 gl) {

    gl.glPushMatrix();
    drawFourBasePatterns(gl);

    gl.glRotatef(90.0f, 0.0f, 0.0f, 1.0f);
    gl.glRotatef(90.0f, 0.0f, 1.0f, 0.0f);
    drawFourBasePatterns(gl);
    
    gl.glRotatef(90.0f, 0.0f, 0.0f, 1.0f);
    gl.glRotatef(90.0f, 0.0f, 1.0f, 0.0f);
    drawFourBasePatterns(gl);
    gl.glPopMatrix();
    
    drawCornerPatterns(gl);
    
  } // drawPattern()
  
  // draw four faces of the pattern on the icosahedron
  static private void drawFourBasePatterns(GL2 gl) {

    gl.glPushMatrix();
    
    drawBasePattern(gl);
    
    gl.glScalef(1.0f, 1.0f, -1.0f);
    drawBasePattern(gl);
    
    gl.glScalef(-1.0f, 1.0f, 1.0f);
    drawBasePattern(gl);
    
    gl.glScalef(1.0f, 1.0f, -1.0f);
    drawBasePattern(gl);
    
    gl.glPopMatrix();
    
  } // drawFourBasePatterns()
  
  // draw eight non-aligned faces of the pattern on the icosahedron
  static private void drawCornerPatterns(GL2 gl) {
    
    gl.glPushMatrix();
    
    drawCornerPattern(gl);

    gl.glScalef(1.0f, -1.0f, 1.0f);
    drawCornerPattern(gl);
    
    gl.glScalef(1.0f, 1.0f, -1.0f);
    drawCornerPattern(gl);
    
    gl.glScalef(1.0f, -1.0f, 1.0f);
    drawCornerPattern(gl);

    gl.glScalef(-1.0f, 1.0f, 1.0f);
    drawCornerPattern(gl);

    gl.glScalef(1.0f, -1.0f, 1.0f);
    drawCornerPattern(gl);
    
    gl.glScalef(1.0f, 1.0f, -1.0f);
    drawCornerPattern(gl);
    
    gl.glScalef(1.0f, -1.0f, 1.0f);
    drawCornerPattern(gl);
    
    gl.glPopMatrix();
    
  } // drawCornerPatterns()
  
  // draw a non-aligned face of the pattern on the icosahedron 
  static private void drawCornerPattern(GL2 gl) {
    
    final float axis[] = Icosahedron.vertex(0);
    final float angle = -360.0f/5;

    gl.glPushMatrix();
    gl.glRotatef(angle, axis[0], axis[1], axis[2]);
    drawBasePattern(gl);
    gl.glPopMatrix();
    
  } // drawCornerPattern()
  
  // draw a single face of the pattern (unit radius)
  static private void drawBasePattern(GL2 gl) {
    
    switch ( kDisplayMethod ) {
      
      case 0: {
        // draw direct
        kEdgeArray.rewind();
        gl.glBegin(GL.GL_LINES);
        for ( int k = 0 ; k < kEdgeArray.limit(); k+=2 ) {
          final int v1 = kEdgeArray.get(),
                    v2 = kEdgeArray.get();
          kVertexArray.position( 3*v1 );
          final float x1 = kVertexArray.get(),
                      y1 = kVertexArray.get(),
                      z1 = kVertexArray.get();
          gl.glVertex3f(x1, y1, z1);
          kVertexArray.position( 3*v2 );
          final float x2 = kVertexArray.get(),
                      y2 = kVertexArray.get(),
                      z2 = kVertexArray.get();
          gl.glVertex3f(x2, y2, z2);
        }
        gl.glEnd();
      } break;
      
      case 1: {
        // draw from vertex arrays
        kEdgeArray.rewind();
        gl.glDrawElements(GL.GL_LINES, 630, //kEdgeArray.limit(), 
                          GL.GL_UNSIGNED_SHORT, kEdgeArray);
      } break;

      case 2: {
        assert(false); // not implemented yet
      } break;
      
      default: {
        assert(false);
      } break;
      
    }
    
  } // drawBasePattern()
  
  // set the current position of the vertex array buffer to the vertex (i,j)
  static private void setVertexArray(int i, int j) {
    
    kVertexArray.position( 3*vertexIndex(i, j) );
    
  } // setVertexArray()
  
  // return the flattened (one-dimensional) index for the the vertex (i,j)
  static private int vertexIndex(int i, int j) {
    
    assert( i >= 0 && i < kNumDivisions );
    assert( j >= 0 && j < kNumDivisions );
    assert( i + j < kNumDivisions );
    
    final int index = i + ((2*kNumDivisions-(j-1))*j)/2;
    assert( index >= 0 && index < (kNumDivisions*(kNumDivisions+1))/2 );

    return index;
    
  } // vertexIndex()

} // class CageGL
