/*
 *  Env.java
 *  Copyright Simon Hern 2009
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package Space;

import java.awt.*;
import java.io.*;
import java.net.*;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.nio.ShortBuffer;
import java.util.*;
import javax.media.opengl.glu.*;
import com.sun.opengl.util.awt.TextRenderer;

// some global variables and functions
public class Env {

  // the component that's in charge
  static private Component kOwner;
  
  // whether to display debug messages, timing statistics
  static private boolean kDebugMode        = false,
                         kPerformanceStats = false;
  
  // size of JOGL canvas (pixels)
  static private int kScreenWidth  = 1,
                     kScreenHeight = 1;
  
  // whether to use OpenGL buffer objects
  static private boolean kUseBufferObjs = true;
  
  // assorted helper objects
  static private Random       kRandom;
  static private KeyMonitor   kKeyMonitor;
  static private MouseMonitor kMouseMonitor;
  //static private Resources    kResources;
  //static private Sounds       kSounds;
  static private GLU          kGlu;
  static private TextRenderer kTextRenderer;
  
  // this sets up a global Env for the application
  static public void initialize(Component owner) {
  
    kOwner = owner;

    kRandom       = new Random();
    kKeyMonitor   = new KeyMonitor(owner);
    kMouseMonitor = new MouseMonitor();
    //kResources  = new Resources();
    //kSounds     = new Sounds();
    kGlu          = null;
    kTextRenderer = null;
    
  } // initialize()

  // whether debug messages, timing statistics should be displayed
  static public boolean debugMode() { return kDebugMode; }
  static public boolean performanceStats() { return kPerformanceStats; }
  static public void setVerbose() { kDebugMode = kPerformanceStats = true; }

  // ratio of screen dimensions (width divided by height)
  static public float aspectRatio() {return kScreenWidth/(float)kScreenHeight;}

  // size of screen (pixels)
  static public void setScreenSize(int w, int h) 
    { kScreenWidth=w; kScreenHeight=h; }
  static public int screenWidth() { return kScreenWidth; }
  static public int screenHeight() { return kScreenHeight; }
  
  // corners of the universe are at (+/-r, +/-r, +/-r)
  static public float worldRadius() { return 7500.0f; }
  
  // return reference to keyboard and mouse monitors
  static public KeyMonitor   keys()  { return kKeyMonitor; }
  static public MouseMonitor mouse() { return kMouseMonitor; }
  
  // return reference to game resources
  //static public Resources resources() { return kResources; }
  
  // return reference to game audio
  //static public Sounds sounds() { return kSounds; }

  // return reference to GLU instance
  static public GLU glu() { return kGlu; }
  static public void setGlu(GLU glu) { kGlu = glu; }

  // return reference to the TextRenderer
  static public TextRenderer textRenderer() { return kTextRenderer; }
  static public void setTextRenderer(TextRenderer t) { kTextRenderer = t; }
  
  // whether to use OpenGL buffer objects (they can crash on some machines)
  static public boolean useBufferObjs() { return kUseBufferObjs; }
  static public void disableBufferObjs() { 
    kUseBufferObjs = false;
    if ( kDebugMode ) System.out.println("Disabling buffer objects");
  }
  
  // assorted functions for returning random numbers
  static public float randomFloat() { // in range [0,1]
    return kRandom.nextFloat(); 
  } // randomFloat()
  static public float randomFloat(float a, float b) { // in range [a,b] 
    return ( a + (b-a)*kRandom.nextFloat() ); 
  } // randomFloat()
  static public double randomDouble() { // in range [0,1]
    return kRandom.nextDouble(); 
  } // randomDouble()
  static public double randomDouble(double a, double b) { // in range [a,b] 
    return ( a + (b-a)*kRandom.nextDouble() ); 
  } // randomDouble()
  static public int randomInt(int n) { // in range [0,n-1] 
    return kRandom.nextInt(n); 
  } // randomInt()
  static public int randomInt(int a, int b) { // in range [a,b]
    if ( a > b )      return ( b + kRandom.nextInt(a-b+1) );
    else if ( a < b ) return ( a + kRandom.nextInt(b-a+1) );
    else              return a;
  } // randomInt()
  static public boolean randomBoolean() { 
    return kRandom.nextBoolean(); 
  } // randomBoolean()

  // assorted modulo-type functions
  static public int fold(int a, int b) {
    // result is between 0 and (b-1)
    if ( a >= 0 ) return (a%b);
    else {
      int temp = b + (a%b);
      return ( (temp==b) ? 0 : temp );
    }
  } // fold()
  static public double fold(double a, double b) {
    // result is in interval [0,b)
    // (probably a more efficient way of doing this?)
    return ( a - b*Math.floor(a/b) );
  } // fold()
  static public float fold(float a, float b) {
    // result is in interval [0,b)
    // (probably a more efficient way of doing this?)
    return ( a - b*(float)Math.floor(a/b) );
  } // fold()
  
  // more modulo-type functions
  static public float foldNearTo(float a, float target, float modSize) {
    return ( target + fold(a-target+0.5f*modSize, modSize) - 0.5f*modSize );
  } // foldNearTo()
  
  // distance moved with maximum deceleration
  static public float stoppingDistance(float currentSpeed, float maxDecel) {
    
    assert( currentSpeed >= 0.0f );
    assert( maxDecel > 0.0f );
    
    // v(t) = v0 - a*t
    // t1 = v0/a
    // d = int[0,t1](v0 - a*t) = [0,t1](v0*t - 0.5*a*t^2) = 0.5*a*t1^2 - v0*t1

    float t1 = currentSpeed/maxDecel;
    return t1*(currentSpeed - 0.5f*maxDecel*t1);
    
  } // stoppingDistance()
  
  // create a direct buffer holding the specified number of floats
  static public FloatBuffer allocateFloatBuffer(int size) {

    final int numBytes = 4*size;
    ByteBuffer buffer = ByteBuffer.allocateDirect(numBytes);
    buffer.order(ByteOrder.nativeOrder());
    return buffer.asFloatBuffer();

  } // allocateFloatBuffer()
  
  // create a direct buffer of the specified number of shorts
  static public ShortBuffer allocateShortBuffer(int size) {

    final int numBytes = 2*size;
    ByteBuffer buffer = ByteBuffer.allocateDirect(numBytes);
    buffer.order(ByteOrder.nativeOrder());
    return buffer.asShortBuffer();

  } // allocateShortBuffer()
  
} // class Env
