/*
 *  Game.java
 *  Copyright Simon Hern 2009
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */
 
package Space;

import Space.planets.*;
import java.util.*;
import javax.media.opengl.*;

// controller of the game
public class Game {

  // light details
  private static final float kLightFirstAngle   = 45.0f,
                             kLightElevations[] = { 45.0f, -30.0f, -45.0f, -70.0f },
                             kLightAmbient[]    = { 0.2f, 0.2f, 0.2f, 1.0f },
                             kLightDiffuse[]    = { 0.8f, 0.8f, 0.8f, 1.0f },
                             kLightSpecular[]   = { 0.0f, 0.0f, 0.0f, 1.0f };
  
  // camera details
  public static final float kCameraLensScale = 0.5f,
                            kCameraOffset    = 1.0f,
                            kCameraNearPlane = kCameraOffset + 0.0f,
                            kCameraFarPlane  = kCameraOffset 
                                             + 3.6f*Env.worldRadius();

  // background details
  private static final float kBackgroundColour[] = {0,0,0};
  
  // width of outlines
  private static final float kLineWidth = 1.2f;
  
  // directions to the light sources
  private float mLightDirections[][] = null;

  // current camera details
  private float mCameraX,
                mCameraY,
                mCameraZ;
  
  // culling planes (normalized, through camera position, point into frustrum)
  private double mCullPlanes[][];
  
  // game objects
  private LinkedList<Sprite> mSprites;
  private Player             mPlayer;
  private HUD                mHUD;
  private Stars              mStars;
  private Cage               mCage;
          
  // constructor
  public Game() {

    // set up lights
    final int numLights = kLightElevations.length;
    mLightDirections = new float[numLights][4];
    for ( int k = 0 ; k < numLights ; k++ ) {
      float phi   = kLightElevations[k];
      float r     = (float)Math.cos(phi*Math.PI/180.0),
            y     = (float)Math.sin(phi*Math.PI/180.0);
      float theta = kLightFirstAngle + (360.0f*k)/numLights;
      float x     = r*(float)Math.cos(theta*Math.PI/180.0),
            z     = r*(float)Math.sin(theta*Math.PI/180.0);
      mLightDirections[k][0] = x;
      mLightDirections[k][1] = y;
      mLightDirections[k][2] = z;
      mLightDirections[k][3] = 0.0f;
    }

    // set up game objects
    mHUD = new HUD();
    mPlayer = new Player(mHUD);
    mStars = new Stars(5748925);
    mSprites = new LinkedList<Sprite>();
    mSprites.add(mPlayer);
    
    mCage = new Cage( Env.worldRadius() );
    mSprites.add(mCage);
        
    mSprites.add(new Planet(-2600.0f, 0.0f, 3200.0f,
                            1000.0f, 8.0f, 
                            25.0f, 160.0f, new CloudySkin()));
    mSprites.add(new Planet(-600.0f, 1000.0f, 3800.0f,
                            400.0f, 10.0f, 
                            50.0f, 260.0f, new RoughGreySkin()));
    mSprites.add(new Planet(100.0f, 3000.0f, 1600.0f,
                            1400.0f, 5.0f, 
                            80.0f, 20.0f, new CellularSkin()));
    mSprites.add(new Planet(1400.0f, -2000.0f, -1800.0f,
                            2400.0f, 3.0f, 
                            45.0f, 45.0f, new FieryOrangeSkin()));

    mSprites.add(new Station(-2600.0f, 0.0f, 3200.0f,
                             1200.0f, 0.0f,
                             25.0f, 160.0f,
                             35.0f, 140.0f));
    mSprites.add(new Station(100.0f, 3000.0f, 1600.0f,
                             1600.0f, 0.0f,
                             80.0f, 20.0f,
                             90.0f, 30.0f));

    for ( int k = 0 ; k < 50 ; k++ ) {
      mSprites.add(new UFO(-1000+500*Env.randomFloat(-1.0f, +1.0f),
                           +2000+500*Env.randomFloat(-1.0f, +1.0f),
                           +3000+500*Env.randomFloat(-1.0f, +1.0f)));
    }
    
    mSprites.add(new WarpGate(2300.0f, 600.0f, -3900.0f));
    
    // initial camera
    positionCamera();
    setPlayerRadius();

  } // constructor
  
  // initialize
  public void init(GL2 gl) {
    
    gl.glClearColor(kBackgroundColour[0], 
                    kBackgroundColour[1],
                    kBackgroundColour[2], 1.0f);
    
    gl.glEnable(GL.GL_CULL_FACE);
    gl.glEnable(GL.GL_DEPTH_TEST);

    gl.glLightModelfv(GL2.GL_LIGHT_MODEL_AMBIENT, kLightSpecular, 0);
    
    gl.glEnable(GL2.GL_LIGHTING);
    gl.glEnable(GL2.GL_LIGHT0);
    gl.glLightfv(GL2.GL_LIGHT0, GL2.GL_AMBIENT, kLightAmbient, 0);
    gl.glLightfv(GL2.GL_LIGHT0, GL2.GL_DIFFUSE, kLightDiffuse, 0);
    gl.glLightfv(GL2.GL_LIGHT0, GL2.GL_SPECULAR, kLightSpecular, 0);
    gl.glDisable(GL2.GL_LIGHT1); //!!!
    gl.glLightfv(GL2.GL_LIGHT1, GL2.GL_DIFFUSE, kLightDiffuse, 0);
    gl.glLightfv(GL2.GL_LIGHT1, GL2.GL_SPECULAR, kLightSpecular, 0);
    gl.glDisable(GL2.GL_LIGHT2); //!!!
    gl.glLightfv(GL2.GL_LIGHT2, GL2.GL_DIFFUSE, kLightDiffuse, 0);
    gl.glLightfv(GL2.GL_LIGHT2, GL2.GL_SPECULAR, kLightSpecular, 0);
    gl.glDisable(GL2.GL_LIGHT3); //!!!
    gl.glLightfv(GL2.GL_LIGHT3, GL2.GL_DIFFUSE, kLightDiffuse, 0);
    gl.glLightfv(GL2.GL_LIGHT3, GL2.GL_SPECULAR, kLightSpecular, 0);
    
    gl.glLightfv(GL2.GL_LIGHT0, GL2.GL_POSITION, mLightDirections[0], 0);
    gl.glLightfv(GL2.GL_LIGHT1, GL2.GL_POSITION, mLightDirections[1], 0);
    gl.glLightfv(GL2.GL_LIGHT2, GL2.GL_POSITION, mLightDirections[2], 0);
    gl.glLightfv(GL2.GL_LIGHT3, GL2.GL_POSITION, mLightDirections[3], 0);

    gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
    gl.glLineWidth(kLineWidth);
    
    gl.glEnable(GL2.GL_NORMALIZE);

    Stars.init(gl);
    Cage.init(gl);
    HUD.init(gl);
    Station.init(gl);
    UFO.init(gl);
    WarpGate.init(gl);
    for ( Sprite sp : mSprites ) {
      if ( sp instanceof Planet ) ((Planet)sp).init(gl);
    }
    
  } // init()

  // set up the camera projection
  // (always called in GL_PROJECTION matrix mode)
  public void configureCamera(GL2 gl) {
  
    float w = kCameraLensScale*Env.aspectRatio(),
          h = kCameraLensScale;
    gl.glFrustum(-w, +w, -h, +h, kCameraNearPlane, kCameraFarPlane);
    
    setPlayerRadius();
    
  } // configureCamera()
  
  // radius for collisions (based on camera details)
  private void setPlayerRadius() {
    
    float w = kCameraLensScale*Env.aspectRatio(),
          h = kCameraLensScale,
          d = kCameraNearPlane - kCameraOffset;
    float r = 1.01f*(float)Math.sqrt(w*w + h*h + d*d);
    mPlayer.setRadius(r);
    
  } // setPlayerRadius()
  
  // set up the camera's position and direction
  // (slightly behind the player, pointing in the same direction)
  private void positionCamera() {
  
    mCameraX = mPlayer.getXPos() - kCameraOffset*mPlayer.getXDir();
    mCameraY = mPlayer.getYPos() - kCameraOffset*mPlayer.getYDir();
    mCameraZ = mPlayer.getZPos() - kCameraOffset*mPlayer.getZDir();
    
    // set up culling planes (through camera position, normal into frustrum)
    if ( mCullPlanes == null ) mCullPlanes = new double[5][3];
    
    mCullPlanes[0][0] = mPlayer.getXDir();
    mCullPlanes[0][1] = mPlayer.getYDir();
    mCullPlanes[0][2] = mPlayer.getZDir();
    
    double x = kCameraLensScale*Env.aspectRatio(),
           y = kCameraLensScale,
           z = kCameraNearPlane;
    double d1 = Math.sqrt(y*y + z*z),
           d2 = Math.sqrt(x*x + z*z);
    
    mCullPlanes[1][0] = 0.0;
    mCullPlanes[1][1] = +z/d1;
    mCullPlanes[1][2] = -y/d1;
      
    mCullPlanes[2][0] = 0.0;
    mCullPlanes[2][1] = -z/d1;
    mCullPlanes[2][2] = -y/d1;
      
    mCullPlanes[3][0] = +z/d2;
    mCullPlanes[3][1] = 0.0;
    mCullPlanes[3][2] = -x/d2;
      
    mCullPlanes[4][0] = -z/d2;
    mCullPlanes[4][1] = 0.0;
    mCullPlanes[4][2] = -x/d2;
      
    for ( int k = 1 ; k < 5 ; k++ ) mPlayer.rotateVector(mCullPlanes[k]);
    
  } // positionCamera()
  
  // true if an object is not visible to the camera
  // (bit sloppy, some non-visible objects won't be culled)
  private boolean cullObject(Sprite s) {
    
    if ( s instanceof Cage )   return false;
    if ( s instanceof Player ) return true;
    
    float dx = s.getXPos() - mCameraX,
          dy = s.getYPos() - mCameraY,
          dz = s.getZPos() - mCameraZ;
    float r  = s.getVisibleRadius();
    
    if ( s instanceof Planet ) {
      // quick hack
      Planet planet = (Planet)s;
      final float d = (float)Math.sqrt(dx*dx + dy*dy + dz*dz);
      final float pixels = (r/d)*kCameraNearPlane*Env.screenHeight(); 
      planet.setPixelRadius(pixels);
    }
    
    for ( int k = 0 ; k < mCullPlanes.length ; k++ ) {
      double dot = mCullPlanes[k][0]*dx 
                 + mCullPlanes[k][1]*dy
                 + mCullPlanes[k][2]*dz;
      if ( k == 0 ) s.setZDepth(dot);
      if ( dot+r < 0.0f ) return true;
    }
    return false;
    
  } // cullObject()
  
  // update positions of objects
  public void advance(float dt) {
  
    for ( Sprite s : mSprites ) {
      if ( s instanceof Player ) continue;
      s.advance(dt, mSprites);
    }
    
    mPlayer.advance(dt, mSprites);
    positionCamera();
  
  } // advance()
  
  // display
  public void display(GL2 gl) {
  
    //float v = Env.aspectRatio() * kCameraLensScale / kCameraNearPlane;
    //mViewWidth = (float)( (180.0f/Math.PI) * Math.atan(v) );
        
    gl.glClear(GL.GL_DEPTH_BUFFER_BIT | GL.GL_COLOR_BUFFER_BIT);

    // camera
    gl.glPushMatrix();
    mPlayer.rotateCamera(gl);

    // stars
    mStars.display(gl);
    
    // camera again
    gl.glTranslatef(-mCameraX, -mCameraY, -mCameraZ);

    // objects
    gl.glLightfv(GL2.GL_LIGHT0, GL2.GL_POSITION, mLightDirections[0], 0);
    gl.glLightfv(GL2.GL_LIGHT1, GL2.GL_POSITION, mLightDirections[1], 0);
    gl.glLightfv(GL2.GL_LIGHT2, GL2.GL_POSITION, mLightDirections[2], 0);
    gl.glLightfv(GL2.GL_LIGHT3, GL2.GL_POSITION, mLightDirections[3], 0);
    displayObjects(gl);
    
    gl.glPopMatrix();

    // head-up display
    enterOrthoMode(gl);
    mHUD.display(gl);
    exitOrthoMode(gl);
    
  } // display()

  // display all objects (without reflection)
  private void displayObjects(GL2 gl) {

    ArrayList<Sprite> transparentObjects = new ArrayList<Sprite>(); 
    
    // display all objects except the player
    int numDrawn = 0,
        numTotal = 0;
    for ( Sprite s : mSprites ) {
      if ( !cullObject(s) ) {
        s.display(gl);
        if ( s.hasTransparency() ) transparentObjects.add(s);
        numDrawn += 1;
      }
      numTotal += 1;
    }
    //System.out.println("drew " + numDrawn + " of " + numTotal);
    
    // sort and draw transparent objects
    Collections.sort(transparentObjects);
    for ( Sprite s : transparentObjects ) {
      s.displayTransparent(gl, mCameraX, mCameraY, mCameraZ);
    }
    
  } // displayObjects()

  // switch to orthographic mode (2D coords -aspect<x<+aspect, -1<y<+1)
  private void enterOrthoMode(GL2 gl) {
  
    gl.glMatrixMode(GL2.GL_PROJECTION);
    gl.glPushMatrix();
    gl.glLoadIdentity();
    Env.glu().gluOrtho2D(-Env.aspectRatio(), +Env.aspectRatio(), -1.0f, 1.0f);
    
    gl.glMatrixMode(GL2.GL_MODELVIEW);
    gl.glPushMatrix();
    gl.glLoadIdentity();

    gl.glDisable(GL.GL_DEPTH_TEST);
    gl.glDisable(GL2.GL_LIGHTING);

  } // enterOrthoMode()

  // switch out of orthographic mode
  private void exitOrthoMode(GL2 gl) {
  
    gl.glPopMatrix();
    gl.glMatrixMode(GL2.GL_PROJECTION);
    gl.glPopMatrix();
    gl.glMatrixMode(GL2.GL_MODELVIEW);
 
    gl.glEnable(GL.GL_DEPTH_TEST);
    gl.glEnable(GL2.GL_LIGHTING);

  } // exitOrthoMode()

} // class Game
 