/*
 *  HUD.java
 *  Copyright Simon Hern 2009
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package Space;

import java.awt.geom.*;
import java.util.*;
import javax.media.opengl.*;

// head-up display (user interface)
public class HUD {

  // details of a target
  private class Target {
    public float mX, mY;
    public float mRadius;
    public float mDistance;
    public Sprite mSprite;
    public Target(float x, float y, float r, float d, Sprite s) 
      { mX=x; mY=y; mRadius=r; mDistance=d; mSprite=s; }
  } // class Target
  
  // appearance of central engine circle
  private static final int   kNumCirclePoints = 40; 
  private static final float kEngineColour[]  = { 0.3f, 0.3f, 1.0f, 1.0f };
  private static final float kEngineRadius    = 0.2f,
                             kEngineMarks     = 0.10f,
                             kEngineSpeed     = 0.06f;
  
  // appearance of target symbol
  private static final int   kNumTargetMarks   = 6;
  private static final float kTargetSymbolMin  = 0.01f,
                             kTargetSymbolMax  = 0.05f;
  private static final float kTargetAngleSpeed = 2.0f;
  private static final float kTargetColour[]   = { 0.3f, 0.3f, 1.0f, 1.0f },
                             kLockOnColour[]   = { 1.0f, 0.8f, 0.0f, 1.0f };
  
  // target radius
  private static final float kTargetRadiusMin    = 0.05f,
                             kTargetRadiusMax    = 0.50f,
                             kTargetRadiusLocked = 0.6f*kTargetRadiusMin;
  
  // attack details
  private static final float kLaserColour[] = { 1.0f, 0.0f, 0.0f, 1.0f };
  private static final float kLaserSeconds  = 0.2f,
                             kLaserOff      = 0.5f;
  private static final float kLaserPosition = 0.01f;
  
  // button appearance
  private static final float kRollButtonSize     = 0.04f,
                             kRollButtonOffset   = 0.06f;
  private static final float kMenuButtonYStart   = 0.4f,
                             kMenuButtonYDelta   = 0.15f,
                             kMenuButtonXSize    = 0.4f,
                             kMenuButtonYSize    = 0.04f;
  private static final float kButtonTextRatio    = 0.8f,
                             kButtonTextBaseline = 0.25f;
  public  static final float kButtonTextHeight   = kButtonTextRatio
                                                   * kMenuButtonYSize;
  private static final int   kMaxVisibleButtons  = 4;
  private static final float kButtonChangeTime   = 0.2f; 
  private static final float kButtonEdgeColour[] = { 1.0f, 1.0f, 1.0f, 1.0f },
                             kButtonGlowColour[] = { 1.0f, 0.8f, 0.0f, 1.0f },
                             kButtonFillColour[] = { 0.3f, 0.3f, 1.0f, 1.0f },
                             kButtonTextColour[] = { 0.2f, 0.2f, 0.2f, 1.0f };
  
  // menu button details
  private static final int kButtonApproach    = 0,
                           kButtonAttack      = 1,
                           kButtonScan        = 2,
                           kButtonCommunicate = 3,
                           kButtonDock        = 4,
                           kButtonWarp        = 5,
                           kNumButtons        = 6;
  private static final boolean kButtonAlive[] = { true, true, false, 
                                                  false, false, false };
  private static final String kButtonNames[][]
                               = { { "Approach", "Stop Approach" },
                                   { "Attack", "Stop Attack" },
                                   { "Scan" },
                                   { "Communicate" },
                                   { "Dock" },
                                   { "Warp" } };
  
  // display lists
  private static int kDisplayListCircle = 0,
                     kDisplayListTarget = 0,
                     kDisplayListRoll1  = 0,
                     kDisplayListRoll2  = 0,
                     kDisplayListButton = 0;

  // current targets
  private LinkedList<Target> mTargets;

  // locked-on target
  private Sprite mLockedOn;
  
  // rotation of target symbol (radians)
  private float mTargetAngle;
  
  // roll button states
  private boolean mRollLeft,
                  mRollRight;
  
  // menu buttons
  private boolean mButtonActive[];
  private int     mButtonState[];
  private float   mButtonTransition[];

  // record of players speed (0.0 to 1.0)
  private float mSpeed;
  
  // laser details
  private float mLaserTimer,
                mLaserX,
                mLaserY;
  
  // prepare display lists
  static public void init(GL2 gl) {
    
    prepareCircle(gl);
    prepareTarget(gl);
    prepareRollButton(gl);
    prepareMenuButton(gl);
    
  } // init()
  
  // constructor
  public HUD() {

    mTargets = new LinkedList<Target>();
    mLockedOn = null;
    mTargetAngle = 0.0f;
  
    mSpeed = 0.0f;

    mLaserTimer = 0.0f;
    mLaserX = mLaserY = 0.0f;
    
    mRollLeft = mRollRight = false;

    mButtonActive     = new boolean[kNumButtons];
    mButtonState      = new int[kNumButtons];
    mButtonTransition = new float[kNumButtons];
    for ( int k = 0 ; k < kNumButtons ; k++ ) {
      mButtonActive[k]     = false;
      mButtonState[k]      = 0;
      mButtonTransition[k] = 0.0f;
    }
    
  } // constructor
  
  // record the players speed (0.0 to 1.0)
  public void setSpeed(float s) { assert(s>=0 && s<=1); mSpeed = s; }
  
  // register a target (screen coordinates)
  // (small targets may be ignored)
  public void setTarget(float x, float y, float r, float d, Sprite s) {
    
    if ( s == mLockedOn ) {
      if ( r <= kTargetRadiusMin && r > kTargetRadiusLocked ) {
        r = kTargetRadiusMin;
      }
    }
    if ( r < kTargetRadiusMin ) return;

    double f = 1.0 - Math.exp(kTargetRadiusMin - r);
    double radius = kTargetRadiusMin + (kTargetRadiusMax-kTargetRadiusMin)*f;
    mTargets.add(new Target(x, y, (float)radius, d, s));
    
  } // setTarget()
  
  // reset the target list
  public void clearTargets() { mTargets.clear(); }

  // what the player is locked on to (null if nothing)
  public Sprite lockedOnTarget() { return mLockedOn; }
  
  // lock on to a new target (and reset menu buttons)
  private void newLockOn(Sprite s) { 
    
    mLockedOn = s; 
    clearButtons();
    
  } // newLockOn()
  
  // lose the player's target lock
  public void clearLockOn() { 
  
    mLockedOn = null; 
    clearButtons();
    
  } // clearLockOn()

  // remove menu buttons
  private void clearButtons() {
    
    for ( int k = 0 ; k < kNumButtons ; k++ ) {
      mButtonActive[k]     = false;
      mButtonState[k]      = 0;
      mButtonTransition[k] = 0.0f;
    }
    
  } // clearButtons()
  
  // the player is approaching the locked-on target
  public boolean approaching() {
    
    if ( mLockedOn != null && mButtonActive[kButtonApproach] ) {
      int state = (int)Math.ceil( mButtonState[kButtonApproach]
                                + mButtonTransition[kButtonApproach] )
                  % kButtonNames[kButtonApproach].length;
      if ( state == 1 ) return true;
    }
    return false;
    
  } // approaching()
  
  // the player is attacking the locked-on target
  public boolean attacking() {
    
    if ( mLockedOn != null && mButtonActive[kButtonAttack] ) {
      int state = (int)Math.ceil( mButtonState[kButtonAttack]
                                + mButtonTransition[kButtonAttack] )
                  % kButtonNames[kButtonAttack].length;
      if ( state == 1 ) return true;
    }
    return false;
    
  } // attacking()
  
  // roll the player
  public boolean rollLeft()  { return mRollLeft; }
  public boolean rollRight() { return mRollRight; }
  
  // react to mouse behaviour
  // (return true if that behaviour concerns the user interface)
  public boolean processMouse(MouseMonitor.State state, boolean clicked) {
    
    if ( state == null || !state.b1 ) {
      assert( !clicked );
      mRollLeft = mRollRight = false;
      return false;
    }

    final int   w    = Env.screenWidth(),
                h    = Env.screenHeight();
    final float x    = (state.x - 0.5f*w)/(0.5f*h),
                y    = (0.5f*h - state.y)/(0.5f*h);
    final float xMax = Env.aspectRatio();
    
    final boolean onLeft  = overButton(x, y, 
                                       -xMax + kRollButtonOffset, 
                                       1.0f - kRollButtonOffset,
                                       kRollButtonSize, kRollButtonSize),
                  onRight = overButton(x, y, 
                                       xMax - kRollButtonOffset, 
                                       1.0f - kRollButtonOffset,
                                       kRollButtonSize, kRollButtonSize);

    // button held
    if ( !clicked ) {
      if ( mRollLeft && onLeft ) {
        return true;
      } else if ( mRollRight && onRight ) {
        return true;
      } else {
        mRollLeft = mRollRight = false;
        return false;
      }
    }

    // roll buttons
    if ( onLeft ) {
      mRollLeft = true;
      return true;
    } else if ( onRight ) {
      mRollRight = true;
      return true;      
    }
    mRollLeft = mRollRight = false;

    // menu buttons
    int visibleButton = -1;
    for ( int k = 0 ; k < kMaxVisibleButtons ; k++ ) {
      if ( overButton(x, y,
                      0.0f, -(kMenuButtonYStart + k*kMenuButtonYDelta),
                      kMenuButtonXSize, kMenuButtonYSize) ) {
        visibleButton = k;
        break;
      }
    }
    int buttonIndex = -1;
    if ( visibleButton >= 0 ) {
      for ( int k = 0 ; k < kNumButtons ; k++ ) {
        if ( mButtonActive[k] ) {
          if ( visibleButton-- == 0 ) {
            buttonIndex = k;
            break;
          }
        }
      }
    }
    if ( buttonIndex >= 0 ) {
      if ( kButtonAlive[buttonIndex] && 
           mButtonTransition[buttonIndex] == 0.0f ) {
        mButtonTransition[buttonIndex] = kButtonChangeTime;
      }
      return true;
    }
    
    // lock on target
    for ( Target t : mTargets ) {
      if ( t.mSprite == mLockedOn ) {
        continue;
      }
      float dx = x - t.mX,
            dy = y - t.mY;
      if ( dx*dx + dy*dy < t.mRadius*t.mRadius ) {
        newLockOn(t.mSprite);
        return true;
      }
    }

    return false;
    
  } // processMouse()
  
  // whether pointer is over button (centre, distance from centre)
  private boolean overButton(float x,       float y, 
                             float xButton, float yButton,
                             float xSize,   float ySize) {
    
    return ( x >= xButton-xSize && x <= xButton+xSize &&
             y >= yButton-ySize && y <= yButton+ySize );
    
  } // overButton()
  
  // time update
  public void advance(float dt) {
    
    mTargetAngle += dt*kTargetAngleSpeed;

    // check that we're locked-on to a valid target
    Target currentTarget = null;
    if ( mLockedOn != null ) {
      for ( Target t : mTargets ) {
        if ( t.mSprite == mLockedOn ) {
          currentTarget = t;
          break;
        }
      }
      if ( currentTarget == null ) clearLockOn();
    }
    
    // set menu buttons
    if ( mLockedOn != null ) {
      float distance = currentTarget.mDistance;
      if ( mLockedOn instanceof Station ) {
        mButtonActive[kButtonApproach] = true;
        mButtonActive[kButtonScan] = true;
        if ( distance < 300.0 ) {
          mButtonActive[kButtonDock] = true;
        } else {
          mButtonActive[kButtonDock]     = false;
          mButtonState[kButtonDock]      = 0;          
          mButtonTransition[kButtonDock] = 0.0f;          
        }
      }
      if ( mLockedOn instanceof UFO ) {
        mButtonActive[kButtonApproach] = true;
        mButtonActive[kButtonAttack] = true;
        mButtonActive[kButtonCommunicate] = true;
      }
      if ( mLockedOn instanceof WarpGate ) {
        mButtonActive[kButtonApproach] = true;
        if ( distance < 600.0 ) {
          mButtonActive[kButtonWarp] = true;
        } else {
          mButtonActive[kButtonWarp]     = false;
          mButtonState[kButtonWarp]      = 0;          
          mButtonTransition[kButtonWarp] = 0.0f;          
        }
      }
    }

    // update menu buttons
    for ( int k = 0 ; k < kNumButtons ; k++ ) {
      if ( mButtonActive[k] && mButtonTransition[k] > 0.0f ) {
        mButtonTransition[k] -= dt;
        if ( mButtonTransition[k] <= 0.0f ) {
          mButtonTransition[k] = 0.0f;
          mButtonState[k] = (mButtonState[k]+1) % kButtonNames[k].length; 
        }
      }
    }

    // update lasers
    if ( mLaserTimer > 0.0f ) mLaserTimer = Math.max(0.0f, mLaserTimer-dt);
    if ( attacking() && mLaserTimer == 0.0f) {
      mLaserTimer = kLaserSeconds;
      mLaserX = kLaserPosition*Env.randomFloat(-1.0f, +1.0f);
      mLaserY = kLaserPosition*Env.randomFloat(-1.0f, +1.0f);
    }
    
  } // advance()
  
  // draw the display (assumed to be in ortho mode)
  public void display(GL2 gl) {
    
    // lasers
    if ( mLaserTimer > kLaserOff*kLaserSeconds ) {
      gl.glDisable(GL2.GL_LIGHTING);
      gl.glEnable(GL.GL_LINE_SMOOTH);
      gl.glEnable(GL.GL_BLEND);
      gl.glColor4fv(kLaserColour, 0);
      gl.glBegin(GL.GL_LINE_STRIP);
        gl.glVertex3f(-1.08f, -1.0f, 0.0f);
        gl.glVertex3f(mLaserX, mLaserY, 0.0f);
        gl.glVertex3f(-0.95f, -1.0f, 0.0f);
      gl.glEnd();
      gl.glBegin(GL.GL_LINE_STRIP);
        gl.glVertex3f(+1.08f, -1.0f, 0.0f);
        gl.glVertex3f(mLaserX, mLaserY, 0.0f);
        gl.glVertex3f(+0.95f, -1.0f, 0.0f);
      gl.glEnd();
      gl.glDisable(GL.GL_BLEND);
      gl.glEnable(GL2.GL_LIGHTING);
    }
    
    // central engine circle
    assert( kDisplayListCircle > 0 );
    gl.glPushMatrix();
    gl.glColor4fv(((mLockedOn!=null) ? kLockOnColour : kEngineColour), 0);
    gl.glScaled(kEngineRadius, kEngineRadius, 1.0f);
    gl.glCallList(kDisplayListCircle);
    
    // speed (around engine circle)
    gl.glDisable(GL2.GL_LIGHTING);
    gl.glEnable(GL.GL_LINE_SMOOTH);
    gl.glEnable(GL.GL_BLEND);
    gl.glBegin(GL.GL_LINE_STRIP);
    float r = 1.0f + kEngineSpeed;
    double thetaMax = 0.5*Math.PI*mSpeed;
    int num = kNumCirclePoints/4;
    for ( int k = 0 ; k < num ; k++ ) {
      double theta = (thetaMax*k)/(num-1) - Math.PI/2;
      gl.glVertex3f(r*(float)Math.cos(theta), r*(float)Math.sin(theta), 0.0f);
    }
    gl.glEnd();
    gl.glDisable(GL.GL_BLEND);
    gl.glEnable(GL2.GL_LIGHTING);
    gl.glPopMatrix();
  
    // targets
    for ( Target t : mTargets ) {
      float col[] = ( (t.mSprite==mLockedOn) ? kLockOnColour : kTargetColour );
      gl.glColor4fv(col, 0);
      gl.glPushMatrix();
      gl.glTranslatef(t.mX, t.mY, 0.0f);
      displayTarget(gl, t.mRadius);
      gl.glPopMatrix();      
    }
    
    // roll buttons
    float x = Env.aspectRatio() - kRollButtonOffset,
          y = 1.0f - kRollButtonOffset;
    
    gl.glPushMatrix();
    gl.glTranslatef(-x, y, 0.0f);
    gl.glScalef(kRollButtonSize, kRollButtonSize, 1.0f);
    gl.glColor4fv((mRollLeft ? kButtonGlowColour : kButtonFillColour), 0);
    gl.glCallList(kDisplayListRoll1);
    gl.glCallList(kDisplayListRoll2);
    gl.glPopMatrix();
    
    gl.glPushMatrix();
    gl.glTranslatef(+x, y, 0.0f);
    gl.glScalef(kRollButtonSize, kRollButtonSize, 1.0f);
    gl.glColor4fv((mRollRight ? kButtonGlowColour : kButtonFillColour), 0);
    gl.glCallList(kDisplayListRoll1);
    gl.glRotatef(180.0f, 0.0f, 0.0f, 1.0f);
    gl.glCallList(kDisplayListRoll2);
    gl.glPopMatrix();
    
    // menu buttons
    float yButton = -kMenuButtonYStart;
    for ( int k = 0 ; k < kNumButtons ; k++ ) {
      if ( mButtonActive[k] ) {
        float col[] = ( mButtonTransition[k] > 0 
                        ? kButtonFillColour : kButtonGlowColour );
        gl.glPushMatrix();
        gl.glTranslatef(0.0f, yButton, 0.0f);
        gl.glColor4fv(col, 0);
        gl.glCallList(kDisplayListButton);
        gl.glPopMatrix();
        displayText(0.0f, yButton, kButtonNames[k][mButtonState[k]]);
        yButton -= kMenuButtonYDelta;
      }
    }

  } // display()
  
  // draw symbols around the target
  private void displayTarget(GL2 gl, float radius) {

    float f = (radius-kTargetRadiusMin)/(kTargetRadiusMax-kTargetRadiusMin);
    float r = kTargetSymbolMin + f*(kTargetSymbolMax-kTargetSymbolMin);

    for ( int k = 0 ; k < kNumTargetMarks ; k++ ) {
      double theta = mTargetAngle + (2.0*Math.PI*k)/kNumTargetMarks;
      float  x = radius*(float)Math.cos(theta),
             y = radius*(float)Math.sin(theta);
      
      gl.glPushMatrix();
      gl.glTranslatef(x, y, 0.0f);
      gl.glScalef(r, r, 1.0f);
      gl.glRotatef((float)(theta*180.0/Math.PI), 0.0f, 0.0f, 1.0f);
      gl.glCallList(kDisplayListTarget);
      gl.glPopMatrix();
    }

  } // displayTarget()

  // display text centred on (x,y)
  private void displayText(float x, float y, String text) {

    final int w = Env.screenWidth(),
              h = Env.screenHeight();
    float x0 = 0.5f*(x/Env.aspectRatio() + 1.0f),
          y0 = 0.5f*((y + (2*kButtonTextBaseline-1)*kMenuButtonYSize) + 1.0f);
    
    Env.textRenderer().beginRendering(w, h);
    Env.textRenderer().setColor(kButtonTextColour[0], kButtonTextColour[1], 
                                kButtonTextColour[2], 1.0f);
    
    Rectangle2D rec = Env.textRenderer().getBounds(text);
    int xPixel = (int)Math.round(x0*w - 0.5f*rec.getWidth()),
        //yPixel = (int)Math.round(y0*h + 0.5f*rec.getHeight() + rec.getY());
        yPixel = (int)Math.round(y0*h);
    
    Env.textRenderer().draw(text, xPixel, yPixel);
    
    Env.textRenderer().endRendering(); 
    
  } // displayText()
  
  // prepare circle display list (centre zero, radius 1)
  static private void prepareCircle(GL2 gl) {
  
    if ( kDisplayListCircle != 0 ) return; // already initialized

    kDisplayListCircle = gl.glGenLists(1);
    gl.glNewList(kDisplayListCircle, GL2.GL_COMPILE);
    gl.glPushMatrix();
    
    gl.glDisable(GL2.GL_LIGHTING);
    gl.glEnable(GL.GL_LINE_SMOOTH);
    gl.glEnable(GL.GL_BLEND);
    
    gl.glBegin(GL.GL_LINE_LOOP);
    for ( int k = 0 ; k < kNumCirclePoints ; k++ ) {
      double theta = (2.0*Math.PI*k)/kNumCirclePoints;
      gl.glVertex3f((float)Math.cos(theta), (float)Math.sin(theta), 0.0f);
    }
    gl.glEnd();

    gl.glBegin(GL.GL_LINES);
      gl.glVertex3f(+1.0f, 0.0f, 0.0f);
      gl.glVertex3f(+(1.0f+kEngineMarks), 0.0f, 0.0f);
      gl.glVertex3f(0.0f, -1.0f, 0.0f);
      gl.glVertex3f(0.0f, -(1.0f+kEngineMarks), 0.0f);
    gl.glEnd();
    
    gl.glDisable(GL.GL_BLEND);
    gl.glEnable(GL2.GL_LIGHTING);
    
    gl.glPopMatrix();
    gl.glEndList();

    assert( kDisplayListCircle > 0 );
  
  } // prepareCircle()
    
  // prepare target display list (triangle, radius 1)
  static private void prepareTarget(GL2 gl) {
  
    if ( kDisplayListTarget != 0 ) return; // already initialized

    kDisplayListTarget = gl.glGenLists(1);
    gl.glNewList(kDisplayListTarget, GL2.GL_COMPILE);
    gl.glPushMatrix();
    
    gl.glDisable(GL2.GL_LIGHTING);
    gl.glEnable(GL.GL_LINE_SMOOTH);
    gl.glEnable(GL.GL_BLEND);
    
    gl.glBegin(GL.GL_LINE_LOOP);
    final int num = 3;
    for ( int k = 0 ; k < num ; k++ ) {
      double theta = (2.0*Math.PI*k)/num + Math.PI;
      gl.glVertex3f((float)Math.cos(theta), (float)Math.sin(theta), 0.0f);
    }
    gl.glEnd();

    gl.glDisable(GL.GL_BLEND);
    gl.glEnable(GL2.GL_LIGHTING);
    
    gl.glPopMatrix();
    gl.glEndList();

    assert( kDisplayListTarget > 0 );
  
  } // prepareTarget()
    
  // prepare roll button display list (square with triangle, radius 1)
  static private void prepareRollButton(GL2 gl) {
  
    if ( kDisplayListRoll1 != 0 ) return; // already initialized

    // filled square
    
    kDisplayListRoll1 = gl.glGenLists(1);
    gl.glNewList(kDisplayListRoll1, GL2.GL_COMPILE);
    gl.glPushMatrix();
    
    gl.glDisable(GL2.GL_LIGHTING);
    gl.glEnable(GL.GL_BLEND);
    gl.glShadeModel(GL2.GL_FLAT);

    gl.glBegin(GL2.GL_QUADS);
      gl.glVertex3f(+1.0f, +1.0f, 0.0f);
      gl.glVertex3f(-1.0f, +1.0f, 0.0f);
      gl.glVertex3f(-1.0f, -1.0f, 0.0f);
      gl.glVertex3f(+1.0f, -1.0f, 0.0f);
    gl.glEnd();
    
    gl.glDisable(GL.GL_BLEND);
    gl.glEnable(GL2.GL_LIGHTING);
    
    gl.glPopMatrix();
    gl.glEndList();

    // outline
    
    kDisplayListRoll2 = gl.glGenLists(1);
    gl.glNewList(kDisplayListRoll2, GL2.GL_COMPILE);
    gl.glPushMatrix();
    
    gl.glColor4fv(kButtonEdgeColour, 0);

    gl.glDisable(GL2.GL_LIGHTING);
    gl.glEnable(GL.GL_LINE_SMOOTH);
    gl.glEnable(GL.GL_BLEND);

    gl.glBegin(GL.GL_LINE_LOOP);
      gl.glVertex3f(+1.0f, +1.0f, 0.0f);
      gl.glVertex3f(+1.0f, -1.0f, 0.0f);
      gl.glVertex3f(-1.0f, -1.0f, 0.0f);
      gl.glVertex3f(-1.0f, +1.0f, 0.0f);
    gl.glEnd();

    final float r = 0.7f;
    final float x = 0.08f;
    gl.glBegin(GL.GL_LINE_LOOP);
    final int num = 3;
    for ( int k = 0 ; k < num ; k++ ) {
      double theta = (2.0*Math.PI*k)/num + Math.PI;
      gl.glVertex3f(r*(float)Math.cos(theta) + x, 
                    r*(float)Math.sin(theta), 0.0f);
    }
    gl.glEnd();

    gl.glDisable(GL.GL_BLEND);
    gl.glEnable(GL2.GL_LIGHTING);
    
    gl.glPopMatrix();
    gl.glEndList();

    assert( kDisplayListRoll1 > 0 && kDisplayListRoll2 > 0);
  
  } // prepareRollButton()
    
  // prepare menu button display list (rectangle, centred on origin)
  static private void prepareMenuButton(GL2 gl) {
  
    if ( kDisplayListButton != 0 ) return; // already initialized

    kDisplayListButton = gl.glGenLists(1);
    gl.glNewList(kDisplayListButton, GL2.GL_COMPILE);
    gl.glPushMatrix();
    
    gl.glDisable(GL2.GL_LIGHTING);
    gl.glEnable(GL.GL_LINE_SMOOTH);
    gl.glEnable(GL.GL_BLEND);
    gl.glShadeModel(GL2.GL_FLAT);

    gl.glBegin(GL2.GL_QUADS);
      gl.glVertex3f(+kMenuButtonXSize, +kMenuButtonYSize, 0.0f);
      gl.glVertex3f(-kMenuButtonXSize, +kMenuButtonYSize, 0.0f);
      gl.glVertex3f(-kMenuButtonXSize, -kMenuButtonYSize, 0.0f);
      gl.glVertex3f(+kMenuButtonXSize, -kMenuButtonYSize, 0.0f);
    gl.glEnd();
    
    gl.glColor4fv(kButtonEdgeColour, 0);

    gl.glBegin(GL.GL_LINE_LOOP);
      gl.glVertex3f(+kMenuButtonXSize, +kMenuButtonYSize, 0.0f);
      gl.glVertex3f(-kMenuButtonXSize, +kMenuButtonYSize, 0.0f);
      gl.glVertex3f(-kMenuButtonXSize, -kMenuButtonYSize, 0.0f);
      gl.glVertex3f(+kMenuButtonXSize, -kMenuButtonYSize, 0.0f);
    gl.glEnd();

    gl.glDisable(GL.GL_BLEND);
    gl.glEnable(GL2.GL_LIGHTING);
    
    gl.glPopMatrix();
    gl.glEndList();

  } // prepareMenuButton()
    
} // class HUD
