/*
 *  Icosahedron.java
 *  Copyright Simon Hern 2009
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package Space;

// basic geometry details for an icosahedron (vertices on the unit sphere)
public class Icosahedron {

  // constants for an icosahedron
  private static final float kPhi = (1.0f + (float)Math.sqrt(5.0f))/2.0f,
                             kR   = (float)Math.sqrt(1.0f + kPhi*kPhi);
  private static final float kA   = kPhi/kR,
                             kB   = 1.0f/kR;
  
  // coordinates of vertices (unit distance from origin)
  private static final float kVertices[][] = { { +kA, +kB,  0 },
                                               { +kA, -kB,  0 },
                                               { -kA, +kB,  0 },
                                               { -kA, -kB,  0 },
                                               {  0, +kA, +kB },
                                               {  0, +kA, -kB },
                                               {  0, -kA, +kB },
                                               {  0, -kA, -kB },
                                               { +kB,  0, +kA },
                                               { -kB,  0, +kA },
                                               { +kB,  0, -kA },
                                               { -kB,  0, -kA } };
  
  // planes of faces (normals point inwards)
  private static Plane kPlanes[] = { buildPlane(0,  1,  8),
                                     buildPlane(0,  1, 10),
                                     buildPlane(0,  4,  5),
                                     buildPlane(0,  4,  8),
                                     buildPlane(0,  5, 10),
                                     buildPlane(1,  6,  7),
                                     buildPlane(1,  6,  8),
                                     buildPlane(1,  7, 10),
                                     buildPlane(2,  3,  9),
                                     buildPlane(2,  3, 11),
                                     buildPlane(2,  4,  5),
                                     buildPlane(2,  4,  9),
                                     buildPlane(2,  5, 11),
                                     buildPlane(3,  6,  7),
                                     buildPlane(3,  6,  9),
                                     buildPlane(3,  7, 11),
                                     buildPlane(4,  8,  9),
                                     buildPlane(5, 10, 11),
                                     buildPlane(6,  8,  9),
                                     buildPlane(7, 10, 11) };

  // number of vertices 
  public static int numVertices() { return kVertices.length; }
  
  // read-only access to a vertex
  public static float[] vertex(int index) {
    
    assert( index >= 0 && index < kVertices.length );
    return kVertices[index];
    
  } // vertex()
  
  // number of face planes 
  public static int numPlanes() { return kPlanes.length; }
  
  // read-only access to a face plane
  public static Plane plane(int index) {
    
    assert( index >= 0 && index < kPlanes.length );
    return kPlanes[index];
    
  } // plane()
  
  // returns a copy of the face planes scaled to the specified radius
  public static Plane[] getPlanes(float radius) { 
    
    assert( radius > 0.0f );
    
    Plane planes[] = new Plane[kPlanes.length];
    for ( int k = 0 ; k < kPlanes.length ; k++ ) {
      float n[] = kPlanes[k].normal();
      planes[k] = new Plane(n[0], n[1], n[2], radius*kPlanes[k].offset());
    }
    
    return planes; 
    
  } // getPlanes()
  
  // a plane through three vertices, normal towards the origin
  private static Plane buildPlane(int i1, int i2, int i3) {

    assert( i1 >= 0 && i1 < kVertices.length );
    assert( i2 >= 0 && i2 < kVertices.length );
    assert( i3 >= 0 && i3 < kVertices.length );
    
    final float v1[] = kVertices[i1],
                v2[] = kVertices[i2],
                v3[] = kVertices[i3];
    
    final float dA[] = new float[]{ v2[0]-v1[0], v2[1]-v1[1], v2[2]-v1[2] },
                dB[] = new float[]{ v3[0]-v1[0], v3[1]-v1[1], v3[2]-v1[2] };
    
    float n[] = new float[]{ dA[1]*dB[2] - dA[2]*dB[1],
                             dA[2]*dB[0] - dA[0]*dB[2],
                             dA[0]*dB[1] - dA[1]*dB[0] };
    
    final float d2 = n[0]*n[0] + n[1]*n[1] + n[2]*n[2];
    assert( d2 > 1.0e-6 );
    final float d = (float)Math.sqrt(d2); 
    for ( int k = 0 ; k < 3 ; k++ ) n[k] /= d;
    
    float offset = n[0]*v1[0] + n[1]*v1[1] + n[2]*v1[2];
    if ( offset > 0 ) {
      for ( int k = 0 ; k < 3 ; k++ ) n[k] = -n[k];
      offset = -offset;
    }

    return new Plane(n[0], n[1], n[2], offset);
    
  } // buildPlane()
    
} // class Icosahedron
