/*
 *  Planet.java
 *  Copyright Simon Hern 2009
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package Space;

import Space.planets.*;
import java.util.LinkedList;
import javax.media.opengl.*;

// a large, spinning sphere
public class Planet extends Sprite {

  // size of the atmospheric effect around the planet
  private static final float kHaloScale = 1.1f;
  
  // extra bit on the bounding sphere
  private static final float kExtraRadius = 10.0f;

  // size of texture to use for planet surface
  private static final int kTextureSize = 512;
  
  // planet size
  private float mRadius,
                mRadiusWithHalo,
                mRadiusBounding;
  
  // position of planet centre
  private float mXPos,
                mYPos,
                mZPos;
  
  // rotation details (degrees, degrees per second)
  private float mAngle;
  private float mAngSpeed;
  
  // orientation (degrees, rotation axis relative to y axis)
  private float mTiltAngle;
  private float mXTilt, 
                mZTilt;

  // textured sphere to display
  private Globe mGlobe;
  
  // prepare resources
  public void init(GL2 gl) {
  
    mGlobe.prepare(gl);
    
  } // init()
  
  // discard resources
  public void dispose(GL2 gl) {
    
    mGlobe.dispose(gl);
    
  } // dispose()
  
  // constructor (both tilt factors in degrees)
  public Planet(float xPos, float yPos, float zPos,
                float radius, float angSpeed,
                float tiltAngle, float tiltDirection,
                SphericalSkin skin) {
  
    mXPos = xPos;
    mYPos = yPos;
    mZPos = zPos;
    
    mRadius         = radius;
    mRadiusWithHalo = mRadius*kHaloScale;
    mRadiusBounding = mRadiusWithHalo + kExtraRadius;
    
    mAngle    = Env.randomFloat(0.0f, 360.0f);
    mAngSpeed = angSpeed;

    mTiltAngle = tiltAngle;
    mXTilt = (float)Math.cos(tiltDirection*Math.PI/180.0f);
    mZTilt = (float)Math.sin(tiltDirection*Math.PI/180.0f);

    mGlobe = new Globe(skin, kTextureSize);
    
  } // constructor
  
  // position (centre of bounding sphere)
  @Override public float getXPos() { return mXPos; }
  @Override public float getYPos() { return mYPos; }
  @Override public float getZPos() { return mZPos; }

  // radii of spheres 
  @Override public float getCollisionRadius() { return mRadiusBounding; } 
  @Override public float getVisibleRadius() { return mRadiusWithHalo; }
  @Override public float getTargetRadius() { return 0.0f; }
  
  // the halo effect around the planet uses transparency
  @Override
  public boolean hasTransparency() { return true; }
  
  // change position, etc.
  @Override
  public void advance(float dt, LinkedList<Sprite> sprites) {
    
    mAngle += dt*mAngSpeed;
    
  } // Sprite.advance()
  
  // size of the planet on screen (may be approximate)
  public void setPixelRadius(float radius) {
    
    mGlobe.setDepthFromSize(radius);
    
  } // setPixelRadius()
  
  // draw the planet (but not the transparent halo)
  @Override
  public void display(GL2 gl) {
    
    gl.glPushMatrix();
    gl.glTranslatef(mXPos, mYPos, mZPos);
    gl.glScalef(mRadius, mRadius, mRadius);
    gl.glRotatef(mTiltAngle, mXTilt, 0.0f, mZTilt);
    gl.glRotatef(-mAngle, 0.0f, 1.0f, 0.0f);
    mGlobe.display(gl);
    gl.glPopMatrix();
    
  } // Sprite.display()

  // draw the atmospheric halo around the planet
  @Override
  public void displayTransparent(GL2 gl, float cameraX, 
                                 float cameraY, float cameraZ) {
    
    float viewXyz[] = {(cameraX-mXPos)/mRadius, 
                       (cameraY-mYPos)/mRadius, 
                       (cameraZ-mZPos)/mRadius };
    
    gl.glPushMatrix();
    gl.glTranslatef(mXPos, mYPos, mZPos);
    gl.glScalef(mRadius, mRadius, mRadius);
    mGlobe.displayHalo(gl, kHaloScale, viewXyz);
    gl.glPopMatrix();
    
  } // Sprite.displayTransparent()

} // class Planet
