/*
 *  Player.java
 *  Copyright Simon Hern 2009
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package Space;

import java.util.LinkedList;
import javax.media.opengl.*;

// the player's "ship" details
public class Player extends Sprite {

  // linear motion (per sec)
  private static final float kMidSpeed = 20.0f,
                             kMaxSpeed = 400.0f,
                             kAccel    = 200.0f,
                             kAccelLow = 20.0f;
  
  // how fast the camera can turn (radians per second)
  private static final float kMaxAngSpeed = 1.2f;

  // how fast the turn direction can change (relative to turn circle, per sec)
  private static final float kMinTurnAccel = 0.3f,
                             kMaxTurnAccel = 1.8f;
  
  // how fast the player rolls (radians, per second)
  private static final float kMaxRollSpeed = 1.5f,
                             kRollAccel    = 5.0f,
                             kRollDecel    = 3.0f;
  
  // how much of the screen the turn and engine circles cover
  private static final float kTurnCircleRadius   = 0.8f,
                             kEngineCircleRadius = 0.2f;
  
  // enumeration of mouse states
  private static final int kMouseUnclicked = 0,
                           kMouseInvalid   = 1,
                           kMouseNavigate  = 2,
                           kMouseInHUD     = 3,
                           kNumMouseStates = 4;
  
  // preferred distance for hostiles
  private static final float kTargetDistanceUfo = 300.0f;
  
  // size
  private float mRadius;
  
  // position
  private float mXPos,
                mYPos,
                mZPos;
  
  // orientation (rotation from standard axes)
  private Quaternion mOrientation;
  
  // current direction (unit vector)
  private float mXDir,
                mYDir,
                mZDir;
  
  // requested turn (y-axis is upwards, (x,y) is point inside unit circle)
  private float mXTurn,
                mYTurn;
  
  // speed of roll (radians per second)
  private float mRollSpeed;
  
  // angular speed when locked on 
  private float mLockOnSpeed;
  
  // engine status
  private boolean mEngineOn;
  private float   mSpeed;

  // current mouse state
  private int mMouseState;
  
  // head-up display
  private HUD mHUD;
  
  // constructor
  public Player(HUD hud) {
  
    mHUD = hud;
    
    mRadius = 0.0f;
    
    mXPos = 0.0f;
    mYPos = 0.0f;
    mZPos = 4500.0f;
  
    mOrientation = new Quaternion();
    updateDirection();
    
    mXTurn = mYTurn = 0.0f;
    mLockOnSpeed = 0.0f;
    
    mEngineOn = false;
    mSpeed = 0.0f;

    mMouseState = kMouseInvalid;
    
  } // constructor
  
  // accessors
  public float getXPos() { return mXPos; }
  public float getYPos() { return mYPos; }
  public float getZPos() { return mZPos; }
  public float getXDir() { return mXDir; }
  public float getYDir() { return mYDir; }
  public float getZDir() { return mZDir; }

  // radius of bounding sphere 
  public void setRadius(float r) { mRadius = r; }
  public float getCollisionRadius() { return mRadius; }
  public float getVisibleRadius() { return mRadius; }
  public float getTargetRadius() { return 0.0f; }
  
  // make the direction vector consistent with the orientation
  private void updateDirection() {
    
    mOrientation.normalize();
    
    double vec[] = { 0.0, 0.0, -1.0 };
    mOrientation.rotate(vec);
    mXDir = (float)vec[0];
    mYDir = (float)vec[1];
    mZDir = (float)vec[2];
    
  } // updateDirection()
  
  // apply rotation from player's coords to global coords
  public void rotateVector(double[] vec) {
    
    assert( vec.length == 3 );
    mOrientation.rotate(vec);
    
  } // rotateVector()
  
  // apply camera rotation
  public void rotateCamera(GL2 gl) {
    
    float theta = (float)(mOrientation.getAngle()*180.0/Math.PI);
    double axis[] = new double[3];
    mOrientation.getAxis(axis);
    gl.glRotatef(-theta, (float)axis[0], (float)axis[1], (float)axis[2]);

  } // rotateCamera()
  
  //
  public Quaternion getOrientation() { return mOrientation; }
  
  // change position, etc.
  public void advance(float dt, LinkedList<Sprite> sprites) {
  
    MouseMonitor.State state = Env.mouse().getState();
    
    // process the mouse behaviour
    switch ( mMouseState ) {
    
      case kMouseInvalid: {
        if ( state != null && !state.b1 ) mMouseState = kMouseUnclicked;
      } break;

      case kMouseUnclicked: {
        if ( state == null ) {
          mMouseState = kMouseInvalid;
        } else if ( state.b1 ) {
          final boolean inHUD = mHUD.processMouse(state, true);
          if ( inHUD ) {
            mMouseState = kMouseInHUD;
          } else {
            mMouseState = kMouseNavigate;
            navigate(state.x, state.y, true, dt);
          }
        }
      } break;
      
      case kMouseNavigate: {
        if ( state == null ) {
          mMouseState = kMouseInvalid;
        } else if ( !state.b1 ) {
          mMouseState = kMouseUnclicked;
        } else {
          navigate(state.x, state.y, false, dt);
        }
      } break;
      
      case kMouseInHUD: {
        final boolean inHUD = mHUD.processMouse(state, false);
        if ( !inHUD ) mMouseState = kMouseInvalid;
      } break;
      
      default: {
        assert(false);
      } break;
    
    }
    if ( mMouseState != kMouseNavigate ) {
      mEngineOn = false;
      updateTurnTarget(0.0f, 0.0f, dt);      
    }
    
    // update direction and position
    turnAndMove(dt);
    restrictPosition(sprites);
    
    // update HUD
    updateTargets(sprites);
    mHUD.setSpeed(mSpeed/kMaxSpeed);
    mHUD.advance(dt);
    
  } // Sprite.advance()

  // turn the camera, trigger the engines
  private void navigate(int xPixel, int yPixel, boolean clicked, float dt) {
    
    final int   w  = Env.screenWidth(),
                h  = Env.screenHeight();
    final float r1 = 0.5f*Math.min(w, h)*kEngineCircleRadius,
                r2 = 0.5f*Math.min(w, h)*kTurnCircleRadius;
    final float x  = (xPixel - 0.5f*w),
                y  = (0.5f*h - yPixel);
    float xTarget = x/r2,
          yTarget = y/r2;
    if ( xTarget*xTarget + yTarget*yTarget > 1.0f ) {
      final float d = (float)Math.hypot(xTarget, yTarget);
      xTarget /= d;
      yTarget /= d;
    }
    if ( x*x + y*y <= r1*r1 && clicked ) {
      mEngineOn = true;
    }
    if ( mHUD.lockedOnTarget() != null && x*x + y*y > r1*r1 && clicked ) {
      mHUD.clearLockOn();
    }

    updateTurnTarget(xTarget, yTarget, dt);

  } // navigate()

  // accelerate change of direction
  private void updateTurnTarget(float xTarget, float yTarget, float dt) {
    
    float dx = xTarget - mXTurn,
          dy = yTarget - mYTurn;
    float d = (float)Math.hypot(dx, dy);
    if ( d > 0.0 ) {
      float acc = dt*( kMinTurnAccel 
                     + (kMaxTurnAccel - kMinTurnAccel)
                       *(float)Math.max(Math.hypot(mXTurn, mYTurn),
                                        Math.hypot(xTarget, yTarget)) );
      if ( d > acc ) {
        mXTurn += acc*dx/d;
        mYTurn += acc*dy/d;
      } else {
        mXTurn = xTarget;
        mYTurn = yTarget;
      }
    }
    
  } // updateTurnTarget()

  // update direction and position
  private void turnAndMove(float dt) {
  
    // ignore requested direction, and turn towards locked-on target
    
    Sprite target = mHUD.lockedOnTarget();
    float targetDistance = 0.0f;
    if ( target != null ) {
      double delta[] = { target.getXPos() - mXPos,
                         target.getYPos() - mYPos,
                         target.getZPos() - mZPos };

      Quaternion inverse = new Quaternion(mOrientation);
      inverse.conjugate();
      inverse.rotate(delta);
      
      double d = Math.sqrt( delta[0]*delta[0] 
                          + delta[1]*delta[1] 
                          + delta[2]*delta[2] );
      for ( int k = 0 ; k < 3 ; k++ ) delta[k] /= d;
      
      double theta = Math.acos(-delta[2]);

      if ( Env.stoppingDistance(mLockOnSpeed, kMaxTurnAccel) > theta ) {
        mLockOnSpeed = Math.max(mLockOnSpeed - dt*kMaxTurnAccel, 0.0f);
      } else {
        mLockOnSpeed = Math.min(mLockOnSpeed + dt*kMaxTurnAccel, kMaxAngSpeed);
      }
      double angDelta = Math.min(theta, dt*mLockOnSpeed);
      
      double xAxis = +delta[1],
             yAxis = -delta[0];
      double r = Math.sqrt(xAxis*xAxis + yAxis*yAxis);

      Quaternion q = new Quaternion(angDelta, xAxis/r, yAxis/r, 0.0);
      mOrientation = mOrientation.multiply(q);

      mXTurn = mYTurn = 0.0f;
      
      targetDistance = (float)d - ( target.getCollisionRadius() 
                                  + getCollisionRadius() );
    } else {
      mLockOnSpeed = 0.0f;
    }
    
    // update orientation and direction
    
    if ( mXTurn != 0.0f ) {
      Quaternion q = new Quaternion(-kMaxAngSpeed*mXTurn*dt, 0.0, 1.0, 0.0);
      mOrientation = mOrientation.multiply(q);
    }
    if ( mYTurn != 0.0f ) {
      Quaternion q = new Quaternion(+kMaxAngSpeed*mYTurn*dt, 1.0, 0.0, 0.0);
      mOrientation = mOrientation.multiply(q);      
    }
    
    // roll
    if ( mHUD.rollLeft() ) {
      mRollSpeed = Math.max(-kMaxRollSpeed, mRollSpeed-dt*kRollAccel);
    } else if ( mHUD.rollRight() ) {
      mRollSpeed = Math.min(+kMaxRollSpeed, mRollSpeed+dt*kRollAccel);
    } else {
      if ( mRollSpeed > 0.0f ) {
        mRollSpeed = Math.max(0.0f, mRollSpeed-dt*kRollDecel);
      } else if ( mRollSpeed < 0.0f ) {
        mRollSpeed = Math.min(0.0f, mRollSpeed+dt*kRollDecel);
      }
    }
    if ( mRollSpeed != 0.0f ) {
      Quaternion q = new Quaternion(mRollSpeed*dt, 0.0, 0.0, -1.0);
      mOrientation = mOrientation.multiply(q);            
    }

    // update speed
    
    boolean faster = mEngineOn;
    
    if ( mHUD.approaching() ) {
      assert( target != null );
      float d = Env.stoppingDistance(mSpeed, kAccel);
      if ( target instanceof UFO ) d = Math.max(d, kTargetDistanceUfo);
      faster = ( d < targetDistance );
    }
    
    if ( faster ) {
      final float f = Math.min(1.0f, (kMaxSpeed-mSpeed)/(kMaxSpeed-kMidSpeed)); 
      final float accel = kAccelLow + f*(kAccel-kAccelLow);
      mSpeed = Math.min(mSpeed + dt*accel, kMaxSpeed);
    } else {
      mSpeed = Math.max(mSpeed - dt*kAccel, 0.0f);      
    }
    
    // update position
    
    updateDirection();
    mXPos += dt*mSpeed*mXDir;
    mYPos += dt*mSpeed*mYDir;
    mZPos += dt*mSpeed*mZDir;
    
  } // turnAndMove()

  // pick out visible targets
  public void updateTargets(LinkedList<Sprite> sprites) {

    LinkedList<Sprite> planets = new LinkedList<Sprite>();
    for ( Sprite sp : sprites ) {
      if ( sp instanceof Planet ) planets.add(sp);
    }
    
    Quaternion inverse = new Quaternion(mOrientation);
    inverse.conjugate();

    mHUD.clearTargets();
    
    for ( Sprite sp : sprites ) {
      
      if ( sp instanceof Station ||
           sp instanceof UFO ||
           sp instanceof WarpGate ) {
        if ( eclipsed(sp, planets) ) continue;
    
        double delta[] = { sp.getXPos() - mXPos,
                           sp.getYPos() - mYPos,
                           sp.getZPos() - mZPos };
        double distance = Math.sqrt( delta[0]*delta[0] 
                                   + delta[1]*delta[1] 
                                   + delta[2]*delta[2] );

        inverse.rotate(delta);
        delta[2] -= Game.kCameraOffset;
        if ( delta[2] >= -Game.kCameraNearPlane ) continue;

        double xScreen = (-delta[0]/delta[2]) / Game.kCameraLensScale,
               yScreen = (-delta[1]/delta[2]) / Game.kCameraLensScale;
        double rScreen = (sp.getTargetRadius()/distance) 
                         / Game.kCameraLensScale;
        
        mHUD.setTarget((float)xScreen, (float)yScreen, 
                       (float)rScreen, (float)distance, sp);
      }
      
    }
    
  } // updateTargets()

  // true if a planet is between the player and the sprite
  private boolean eclipsed(Sprite s, LinkedList<Sprite> planets) {
    
    float dx = s.getXPos() - mXPos,
          dy = s.getYPos() - mYPos,
          dz = s.getZPos() - mZPos;
    float d2 = dx*dx + dy*dy + dz*dz;
    float d  = (float)Math.sqrt(d2);
    
    for ( Sprite planet : planets ) {
      float px = planet.getXPos() - mXPos,
            py = planet.getYPos() - mYPos,
            pz = planet.getZPos() - mZPos;
      float pr = planet.getVisibleRadius();
      float p2 = px*px + py*py + pz*pz;
      if ( d2 < p2-pr*pr ) continue;
    
      float dot = dx*px + dy*py + dz*pz;
      if ( dot <= 0.0 ) continue;
      
      float p = (float)Math.sqrt(p2);
      float sAng = (float)Math.acos( dot/(d*p) ),
            pAng = (float)Math.asin( pr/p );
      if ( sAng < pAng ) return true;
    }
    
    return false;
    
  } // eclipsed()
  
  // change position to avoid collisions
  private void restrictPosition(LinkedList<Sprite> sprites) {

    Cage cage = null;
    for ( Sprite sp : sprites ) {
      if ( sp instanceof Cage ) {
        cage = (Cage)sp;
        break;
      }
    }
    assert( cage != null );
    
    for ( Plane p : cage.getPlanes() ) {
      float d = p.distance(mXPos, mYPos, mZPos) - 10*mRadius;
      if ( d < 0 ) {
        float n[] = p.normal();
        mXPos += (-d)*n[0];
        mYPos += (-d)*n[1];
        mZPos += (-d)*n[2];
      }
    }
    
    /*
    final float rMax = Env.worldRadius() - mRadius;
    mXPos = Math.min(+rMax, Math.max(-rMax, mXPos));
    mYPos = Math.min(+rMax, Math.max(-rMax, mYPos));
    mZPos = Math.min(+rMax, Math.max(-rMax, mZPos));
    */
    
    for ( Sprite sp : sprites ) {
      
      if ( sp instanceof Planet ||
           sp instanceof Station ||
           sp instanceof UFO ||
           sp instanceof WarpGate ) {
        repelFromSphere(sp.getCollisionRadius(), 
                        sp.getXPos(), sp.getYPos(), sp.getZPos());
      }
      
    }
    
  } // restrictPosition()
  
  // move away from sphere if overlapping
  private void repelFromSphere(float r, float x, float y, float z) {
    
    r += mRadius;
    float dx = mXPos - x,
          dy = mYPos - y,
          dz = mZPos - z;
    float d2 = dx*dx + dy*dy + dz*dz;
    if ( d2 < r*r ) {
      assert( d2 > 1.0e-6 );
      float d = (float)Math.sqrt(d2);
      mXPos = x + r*dx/d;
      mYPos = y + r*dy/d;
      mZPos = z + r*dz/d;
    }
        
  } // repelFromSphere()
  
} // class Player
