/*
 *  Quaternion.java
 *  Copyright Simon Hern 2009
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package Space;

// quick and dirty quaternion implementation
public class Quaternion {

  // components
  private double mW, mX, mY, mZ;
  
  // constructor (no rotation)
  public Quaternion() {
    
    reset();

  } // constructor (default)
  
  // constructor (copy)
  public Quaternion(Quaternion Q) {
    
    mW = Q.mW;
    mX = Q.mX;
    mY = Q.mY;
    mZ = Q.mZ;
    
  } // constructor (copy)
  
  // constructor (from 3 or 4 components)
  public Quaternion(double components[]) {
    
    set(components);
    
  } // constructor (double[])
  
  // constructor (from angle and axis)
  public Quaternion(double theta, double x, double y, double z) {
    
    set(theta, x, y, z);
    
  } // constructor (theta,x,y,z)
  
  // set to no rotation
  public void reset() {
    
    mW = 1.0;
    mX = mY = mZ = 0.0;

  } // reset() 

  // define quaternion from 3 or 4 components
  public void set(double components[]) {
    
    assert( components.length == 3 || components.length == 4 );
    
    mX = components[0];
    mY = components[1];
    mZ = components[2];
    
    mW = ( components.length == 4 ) ? components[3] : 0.0;
    
  } // set (double[])
  
  // define quaternion from angle (radians) and axis
  public void set(double theta, double x, double y, double z) {
    
    final double d2 = (x*x + y*y + z*z);

    if ( d2 < 1.0e-12 ) {
      reset();
      return;
    }

    final double sth = Math.sin(theta/2);
    final double f   = sth/Math.sqrt(d2); 

    mW = Math.cos(theta/2);
    mX = f*x;
    mY = f*y;
    mZ = f*z;    
  
  } // set()
  
  // access to rotation angle (radians)
  public double getAngle() {
    
    return 2.0*Math.acos(mW);
    
  } // getAngle()

  // access to rotation axis (normalized)
  public void getAxis(double vec[]) {
    
    assert( vec.length == 3 );
    final double d2 = (mX*mX + mY*mY + mZ*mZ);
    if ( d2 < 1.0e-12 ) {
      vec[0] = 1.0;
      vec[1] = 0.0;
      vec[2] = 0.0;
    } else {
      final double d = Math.sqrt(d2);
      vec[0] = mX/d;
      vec[1] = mY/d;
      vec[2] = mZ/d;
    }
    
  } // getAxis()
  
  // enforce unit length
  public void normalize() {
    
    final double d2 = (mW*mW + mX*mX + mY*mY + mZ*mZ);
    if ( d2 < 1.0e-12 ) {
      reset();
    } else {
      final double d = Math.sqrt(d2);
      mW /= d;
      mX /= d;
      mY /= d;
      mZ /= d;
    }
    
  } // normalize()
  
  // reverse the rotation
  public void conjugate() {
    
    mX = -mX;
    mY = -mY;
    mZ = -mZ;
    
  } // conjugate()

  // product (returns this * other)
  public Quaternion multiply(Quaternion other) {
    
    Quaternion Q = new Quaternion();
    
    final double w0 = mW,
                 x0 = mX,
                 y0 = mY,
                 z0 = mZ;
    final double w1 = other.mW,
                 x1 = other.mX,
                 y1 = other.mY,
                 z1 = other.mZ;
    
    Q.mW = w0*w1 - x0*x1 - y0*y1 - z0*z1;
    Q.mX = w0*x1 + w1*x0 + (y0*z1 - z0*y1);
    Q.mY = w0*y1 + w1*y0 + (z0*x1 - x0*z1);
    Q.mZ = w0*z1 + w1*z0 + (x0*y1 - y0*x1);
    
    return Q;
    
  } // multiply()
  
  // rotate a 3-vector
  public void rotate(double vec[]) {
    
    assert( vec.length == 3 );
    
    Quaternion Qvec = new Quaternion(vec);
    
    Quaternion Qconj = new Quaternion(this);
    Qconj.conjugate();
    
    Qvec = multiply( Qvec.multiply(Qconj) );
    vec[0] = Qvec.mX;
    vec[1] = Qvec.mY;
    vec[2] = Qvec.mZ;
    
  } // rotate()
  
} // class Quaternion
