/**
 *  Stars.java
 *  Copyright Simon Hern 2009
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package Space;

import java.awt.image.BufferedImage;
import java.nio.FloatBuffer;
import java.util.Random;
import javax.media.opengl.*;
import com.sun.opengl.util.texture.*;
import com.sun.opengl.util.texture.awt.*;

// background stars (point sprites on a sphere centred on the camera)
public class Stars {

  // number of stars in the sky
  static final private int kNumStars = 5000;

  // number of different star sizes
  static final private int kNumStarSizes = 10;
  
  // maximum star size relative to display size
  static final private float kMaxStarSize = 0.03f;
  
  // minimum star size as a fraction of maximum star size
  static final private float kMinStarScale = 0.25f;
  
  // basic star colour
  static final private float kStarColour[] = { 0.5f, 0.5f, 1.0f };
  
  // size of texture holding star image
  static final private int kTextureSize = 16;
  
  // texture used for point sprite stars
  static private Texture kTexture;
  
  // 0 => direct, 1 => arrays, 2 => buffer objects
  static private int kDisplayMethod = 1;

  // radius of the star sphere (must be between the near and far planes)
  static final private float kSphereRadius = 100.0f; 
  
  // (x,y,z) coordinates of stars (on a sphere centred on the camera)
  private FloatBuffer mVertexArray;

  // initialize (prepare a point sprite texture that looks like a star)
  static public void init(GL2 gl) {
    
    if ( kTexture != null ) return;
    
    if ( !Env.useBufferObjs() && kDisplayMethod == 2 ) kDisplayMethod = 1;
    
    BufferedImage image = new BufferedImage(kTextureSize, kTextureSize, 
                                            BufferedImage.TYPE_INT_ARGB);

    final int centre = kTextureSize/2 - 1;
    final int rgb = 0xFFFFFF;
    final float scale = 1.5f;
    final float f0 = 1.0f/((scale+1)*(scale+1));
    
    for ( int ix = 0 ; ix < kTextureSize ; ix++ ) {
      for ( int iy = 0 ; iy < kTextureSize ; iy++ ) {
        final int d2 = (ix-centre)*(ix-centre) + (iy-centre)*(iy-centre);
        final float d = Math.min(1.0f, (float)Math.sqrt(d2)/centre);
        assert( d >= 0.0f && d <= 1.0f );
        
        final float f = 1.0f/((scale*d+1)*(scale*d+1));
        assert( f >= f0 && f <= 1.0f );
        
        final float a = (f - f0)/(1.0f - f0);
        assert( a >= 0.0f && a <= 1.0f );
        
        final int alpha = (int)Math.round(255*a);
        final int argb = (alpha<<24) + rgb;
        image.setRGB(ix, iy, argb);
      }
    }
    
    kTexture = AWTTextureIO.newTexture(image, true);
    
  } // init()

  // discard any resources
  static public void dispose(GL2 gl) {
    
  } // dispose()
  
  // constructor (define pseudo-random star positions)
  public Stars(int seed) {
    
    Random rand = new Random(seed);
    
    mVertexArray = Env.allocateFloatBuffer(3*kNumStars);
    for ( int k = 0 ; k < kNumStars ; ) {
      float x = 2.0f*rand.nextFloat() - 1.0f,
            y = 2.0f*rand.nextFloat() - 1.0f,
            z = 2.0f*rand.nextFloat() - 1.0f;
      final float d2 = x*x + y*y + z*z;
      if ( d2 <= 1.0f && d2 > 0.01f ) {
        final float scale = kSphereRadius/(float)Math.sqrt(d2);
        mVertexArray.put(scale*x);
        mVertexArray.put(scale*y);
        mVertexArray.put(scale*z);
        k++;
      }
    }

  } // constructor
  
  // draw the object (unit radius, centred on origin)
  public void display(GL2 gl) {
    
    final float maxPointSize = Env.screenHeight()*kMaxStarSize;
    
    gl.glDisable(GL2.GL_LIGHTING);
    gl.glDisable(GL.GL_DEPTH_TEST);
    gl.glEnable(GL.GL_BLEND);

    gl.glShadeModel(GL2.GL_FLAT);
    
    bindStars(gl);
    
    kTexture.enable();
    kTexture.bind();

    gl.glTexEnvi(GL2.GL_POINT_SPRITE, GL2.GL_COORD_REPLACE, GL.GL_TRUE);
    gl.glTexEnvf(GL2.GL_TEXTURE_ENV, GL2.GL_TEXTURE_ENV_MODE, 
                 GL2.GL_MODULATE);
    
    gl.glTexParameteri(GL2.GL_TEXTURE_2D, GL2.GL_TEXTURE_MAG_FILTER,
                       GL2.GL_LINEAR);
    gl.glTexParameteri(GL2.GL_TEXTURE_2D, GL2.GL_TEXTURE_MIN_FILTER,
                       GL2.GL_LINEAR_MIPMAP_LINEAR);
    gl.glTexParameteri(GL2.GL_TEXTURE_2D, GL2.GL_TEXTURE_WRAP_S,
                       GL2.GL_CLAMP_TO_EDGE);
    gl.glTexParameteri(GL2.GL_TEXTURE_2D, GL2.GL_TEXTURE_WRAP_T,
                       GL2.GL_CLAMP_TO_EDGE);

    gl.glEnable(GL2.GL_POINT_SPRITE);
    gl.glColor3f(kStarColour[0], kStarColour[1], kStarColour[2]);    

    for ( int size = 0 ; size < kNumStarSizes ; size++ ) {
      final float scale = kMinStarScale 
                        + ((1.0f-kMinStarScale)*size)/(kNumStarSizes-1);
      gl.glPointSize(scale*maxPointSize);
      
      final int starsStart = (size*kNumStars)/kNumStarSizes,
                starsEnd   = ((size+1)*kNumStars)/kNumStarSizes;
      drawStars(gl, starsStart, starsEnd);      
    }

    gl.glPointSize(1.0f);
    gl.glDisable(GL2.GL_POINT_SPRITE);
    
    kTexture.disable();

    unbindStars(gl);
    
    gl.glDisable(GL.GL_BLEND);
    gl.glEnable(GL.GL_DEPTH_TEST);
    gl.glEnable(GL2.GL_LIGHTING);
    
  } // display()
  
  // prepare to draw stars
  private void bindStars(GL2 gl) {
    
    switch ( kDisplayMethod ) {
      
      case 0: {
        // direct display
      } break;
      
      case 1: {
        // use vertex array
        mVertexArray.rewind();
        gl.glEnableClientState(GL2.GL_VERTEX_ARRAY);
        gl.glVertexPointer(3, GL.GL_FLOAT, 0, mVertexArray);
      } break;

      case 2: {
        // use buffer objects
        assert(false); // not implemented yet
      } break;
      
      default: {
        assert(false);
      } break;
      
    }
      
  } // bindStars()
  
  // finished drawing stars
  private void unbindStars(GL2 gl) {
    
    switch ( kDisplayMethod ) {
      
      case 0: {
        // direct display
      } break;
      
      case 1: {
        // use vertex array
        gl.glDisableClientState(GL2.GL_VERTEX_ARRAY);
      } break;

      case 2: {
        // use buffer objects
        assert(false); // not implemented yet
      } break;
      
      default: {
        assert(false);
      } break;
      
    }
      
  } // unbindStars()
  
  // draw a selection of stars (start <= index < end) at a fixed point size
  private void drawStars(GL2 gl, int startIndex, int endIndex) {
    
    final int num = endIndex - startIndex;

    switch ( kDisplayMethod ) {
      
      case 0: {
        // direct display
        mVertexArray.position( 3*startIndex );
        gl.glBegin(GL.GL_POINTS);
        for ( int k = 0 ; k < num ; k++ ) {
          final float x = mVertexArray.get(),
                      y = mVertexArray.get(),
                      z = mVertexArray.get();
          gl.glVertex3f(x, y, z);
        }
        gl.glEnd();
      } break;
      
      case 1: {
        // use vertex array
        gl.glDrawArrays(GL.GL_POINTS, startIndex, num);
      } break;

      case 2: {
        // use buffer objects
        assert(false); // not implemented yet
      } break;
      
      default: {
        assert(false);
      } break;
      
    }
      
  } // drawStars()
  
} // class Stars
