/*
 *  Station.java
 *  Copyright Simon Hern 2009
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package Space;

import java.util.LinkedList;
import javax.media.opengl.*;

// an object in orbit
public class Station extends Sprite {

  // material details
  private static final float kDiffuseColour[]  = { 0.5f, 0.5f, 0.5f, 1.0f },
                             kSpecularColour[] = { 0.8f, 0.8f, 0.8f, 1.0f };
  private static final float kShininess        = 8.0f;
  
  // outline details
  private static final float kLineColour[] = { 0.2f, 0.2f, 0.2f, 1.0f };
  
  // object shape
  private static final float kShapeMinY   = -100.0f,
                             kShapeMaxY   = +60.0f,
                             kShapeRadius = 30.0f,
                             kArmY        = +20.0f,
                             kArmRadius   = 8.0f,
                             kArmLength   = 70.0f;;
  
  // motion
  private static final float kAngSpeed   = 50.0f,
                             kOrbitSpeed = 100.0f;
  
  // bounding sphere
  private static final float kRadius = 
      1.1f*(float)Math.sqrt(Math.max( 
          kArmY*kArmY + kArmLength*kArmLength,
          kShapeRadius*kShapeRadius + kShapeMinY*kShapeMinY));
  
  // display list
  private static int kDisplayList = 0;

  // orbit details
  private float mOrbitXPos,
                mOrbitYPos,
                mOrbitZPos,
                mOrbitRadius,
                mOrbitAngle,
                mOrbitAngSpeed;
  private Quaternion mOrbitTilt;
  
  // position (of centre of station)
  private float mXPos,
                mYPos,
                mZPos;
  
  // rotation
  private float mAngle;
  
  // orientation (rotation axis relative to y axis)
  private float mTiltAngle;
  private float mXTilt, 
                mZTilt;

  // prepare display list
  static public void init(GL2 gl) {
  
    if ( kDisplayList != 0 ) return; // already initialized

    kDisplayList = gl.glGenLists(1);
    gl.glNewList(kDisplayList, GL2.GL_COMPILE);
    gl.glPushMatrix();
    
    gl.glMaterialfv(GL.GL_FRONT, GL2.GL_AMBIENT_AND_DIFFUSE, kDiffuseColour, 0);
    gl.glMaterialfv(GL.GL_FRONT, GL2.GL_SPECULAR, kSpecularColour, 0);
    gl.glMaterialf(GL.GL_FRONT, GL2.GL_SHININESS, kShininess);

    gl.glShadeModel(GL2.GL_FLAT);
    gl.glPolygonOffset(1.0f, 1.0f);
    gl.glEnable(GL.GL_POLYGON_OFFSET_FILL);
    addShapes(gl, false);
    gl.glDisable(GL.GL_POLYGON_OFFSET_FILL);

    gl.glDisable(GL2.GL_LIGHTING);
    gl.glEnable(GL.GL_LINE_SMOOTH);
    gl.glEnable(GL.GL_BLEND);
    gl.glColor4fv(kLineColour, 0);
    addShapes(gl, true);
    gl.glDisable(GL.GL_BLEND);
    gl.glEnable(GL2.GL_LIGHTING);
    
    gl.glPopMatrix();
    gl.glEndList();

    assert( kDisplayList > 0 );
  
  } // init()
  
  // constructor (angles in degrees)
  public Station(float xOrbit, float yOrbit, float zOrbit,
                 float orbitRadius, float orbitAngle,
                 float orbitTiltAngle, float orbitTiltDirection,
                 float modelTiltAngle, float modelTiltDirection) {
   
    // orbit details
    mOrbitXPos = xOrbit;
    mOrbitYPos = yOrbit;
    mOrbitZPos = zOrbit;
    
    mOrbitRadius = orbitRadius;
    mOrbitAngle = orbitAngle;
    mOrbitAngSpeed = (180.0f/(float)Math.PI)*kOrbitSpeed/mOrbitRadius;

    double x = Math.cos(orbitTiltDirection*Math.PI/180.0f);
    double z = Math.sin(orbitTiltDirection*Math.PI/180.0f);
    mOrbitTilt = new Quaternion(orbitTiltAngle*Math.PI/180.0, x, 0.0, z); 
    
    // orientation of station's rotation axis 
    mTiltAngle = modelTiltAngle;
    mXTilt = (float)Math.cos(modelTiltDirection*Math.PI/180.0f);
    mZTilt = (float)Math.sin(modelTiltDirection*Math.PI/180.0f);
    
    // position of station
    mAngle = Env.randomFloat(0.0f, 360.0f);
    updatePosition();
    
  } // constructor
  
  // accessors
  public float getXPos() { return mXPos; }
  public float getYPos() { return mYPos; }
  public float getZPos() { return mZPos; }
  
  // radii of spheres
  public float getCollisionRadius() { return kRadius; }
  public float getVisibleRadius()   { return kRadius; }
  public float getTargetRadius()    { return kRadius; }
  
  // calculate current position
  private void updatePosition() {
    
    double theta = mOrbitAngle*Math.PI/180.0;
    double vec[] = { Math.cos(theta), 0.0, Math.sin(theta) };
    mOrbitTilt.rotate(vec);
    
    mXPos = mOrbitXPos + mOrbitRadius*(float)vec[0];
    mYPos = mOrbitYPos + mOrbitRadius*(float)vec[1];
    mZPos = mOrbitZPos + mOrbitRadius*(float)vec[2];
    
  } // updatePosition()
  
  // change position, etc.
  public void advance(float dt, LinkedList<Sprite> sprites) {
    
    mAngle += dt*kAngSpeed;
    mOrbitAngle += dt*mOrbitAngSpeed;

    updatePosition();
    
  } // Sprite.advance()
  
  // draw the object
  public void display(GL2 gl) {
    
    assert( kDisplayList > 0 );
  
    gl.glPushMatrix();
    gl.glTranslatef(mXPos, mYPos, mZPos);
    gl.glRotatef(mTiltAngle, mXTilt, 0.0f, mZTilt);
    gl.glRotatef(-mAngle, 0.0f, 1.0f, 0.0f);
    gl.glCallList(kDisplayList);
    gl.glPopMatrix();
    
  } // Sprite.display()

  // triangles and quads that make up the station
  static private void addShapes(GL2 gl, boolean outline) {

    final int numSides = 8;
    float dTheta = 2.0f*(float)Math.PI/numSides;
    for ( int k = 0 ; k < numSides ; k++ ) {
      float theta1 = (k-0.5f)*dTheta,
            theta2 = ( (k == numSides-1) ? -0.5f : (k+0.5f) )*dTheta;
      float x1 = kShapeRadius*(float)Math.cos(theta1),
            z1 = kShapeRadius*(float)Math.sin(theta1),
            x2 = kShapeRadius*(float)Math.cos(theta2),
            z2 = kShapeRadius*(float)Math.sin(theta2);
      addQuad(gl, x1, kShapeMaxY, z1,
                  x2, kShapeMaxY, z2,
                  x2, kShapeMinY, z2,
                  x1, kShapeMinY, z1, outline);
      if ( !outline ) {
        addTriangle(gl,   x2, kShapeMaxY,   z2,
                          x1, kShapeMaxY,   z1,
                        0.0f, kShapeMaxY, 0.0f, false);
        addTriangle(gl,   x1, kShapeMinY,   z1,
                          x2, kShapeMinY,   z2,
                        0.0f, kShapeMinY, 0.0f, false);
      }
    }
    
    float d = kShapeRadius*(float)Math.cos(dTheta/2) + 1.0e-3f;
    float x1 = -kArmRadius,
          x2 = +kArmRadius,
          y1 = kArmY - kArmRadius,
          y2 = kArmY + kArmRadius,
          z1 = d,
          z2 = d + kArmLength;
    for ( int k = 0 ; k < numSides/2 ; k++ ) {
      float xx1=0, xx2=0, zz1=0, zz2=0;
      switch ( k ) {
        case 0: {
          xx1 = +x1;
          xx2 = +x2;
          zz1 = +z1;
          zz2 = +z2;
        } break;
        case 1: {
          xx1 = +z2;
          xx2 = +z1;
          zz1 = -x1;
          zz2 = -x2;
        } break;
        case 2: {
          xx1 = -x1;
          xx2 = -x2;
          zz1 = -z1;
          zz2 = -z2;
        } break;
        case 3: {
          xx1 = -z2;
          xx2 = -z1;
          zz1 = +x1;
          zz2 = +x2;
        } break;
        default: {
          assert(false);
        } break;
      }
      addQuad(gl, xx1, y1, zz1,
                  xx1, y1, zz2,
                  xx1, y2, zz2,
                  xx1, y2, zz1, outline);
      addQuad(gl, xx2, y1, zz1,
                  xx2, y2, zz1,
                  xx2, y2, zz2,
                  xx2, y1, zz2, outline);
      addQuad(gl, xx1, y1, zz1,
                  xx2, y1, zz1,
                  xx2, y1, zz2,
                  xx1, y1, zz2, outline);
      addQuad(gl, xx1, y2, zz1,
                  xx1, y2, zz2,
                  xx2, y2, zz2,
                  xx2, y2, zz1, outline);
      addQuad(gl, xx1, y1, zz1,
                  xx1, y2, zz1,
                  xx2, y2, zz1,
                  xx2, y1, zz1, outline);
      addQuad(gl, xx1, y1, zz2,
                  xx2, y1, zz2,
                  xx2, y2, zz2,
                  xx1, y2, zz2, outline);
    }
    
  } // addShapes()
  
} // class Station
