/*
 *  UFO.java
 *  Copyright Simon Hern 2009
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package Space;

import java.util.LinkedList;
import javax.media.opengl.*;

// a small, flocking object
public class UFO extends Sprite {

  // material details
  private static final float kDiffuseColour[]  = { 0.5f, 0.5f, 0.5f, 1.0f },
                             kSpecularColour[] = { 0.8f, 0.8f, 0.8f, 1.0f };
  private static final float kShininess        = 8.0f;
  
  // outline details
  private static final float kLineColour[] = { 0.2f, 0.2f, 0.2f, 1.0f };
  
  // object shape
  private static final float kMinZ   = -15.0f,
                             kMaxZ   =  +5.0f,
                             kMaxX   = +10.0f,
                             kMaxY   =  +5.0f;
  private static final float kRadius = Math.abs(kMinZ);
  
  // movement
  private static final float kMaxSpeed = 300.0f,
                             kMaxAccel = 200.0f;
  private static final float kStoppingDistance = 
                               Env.stoppingDistance(kMaxSpeed, kMaxAccel);
  
  // repulsion from other objects
  private static final float kCageRange        = 100.0f,
                             kCageHardness     = 1.0f,
                             kPlanetRange      = 200.0f,
                             kPlanetHardness   = 1.0f,
                             kPlayerRange      = 10.0f,
                             kPlayerHardness   = 0.0f,
                             kStationRange     = 300.0f,
                             kStationHardness  = 0.0f,
                             kUfoRange         = 30.0f,
                             kUfoHardness      = 0.0f,
                             kWarpGateRange    = 200.0f,
                             kWarpGateHardness = 1.0f;
  
  // aim for common position and direction
  private static final float kHerdAccel  = 50.0f,
                             kFlockAccel = 10.0f,
                             kFlockRange = 1000.0f;
  private static final int   kNumFlocks  = 5;

  // time delays till change of flock (seconds)
  private static final float kSwitchMinDelay = 60.0f,
                             kSwitchMaxDelay = 300.0f;
  
  // rotational speed
  private static final float kAngSpeedMin   = 10.0f,
                             kAngSpeedMax   = 100.0f;
  private static final float kAngSpeedDelta = 50.0f,
                             kAngDelayMin  = 3.0f,
                             kAngDelayMax  = 10.0f;
  
  // display list
  private static int kDisplayList = 0;

  // position
  private float mXPos,
                mYPos,
                mZPos;

  // velocity
  private float mXVel,
                mYVel,
                mZVel;
  
  // which flock we belong to
  private int mFlockId;

  // how long till change of flock
  private float mSwitchTime;
  
  // rotation
  private float mAngle,
                mAngSpeed,
                mAngSpeedTarget;
  private float mAngSpeedDelay;
  
  // direction
  private Quaternion mDirection;

  // prepare display list
  static public void init(GL2 gl) {
  
    if ( kDisplayList != 0 ) return; // already initialized

    kDisplayList = gl.glGenLists(1);
    gl.glNewList(kDisplayList, GL2.GL_COMPILE);
    gl.glPushMatrix();
    
    gl.glMaterialfv(GL.GL_FRONT, GL2.GL_AMBIENT_AND_DIFFUSE, kDiffuseColour, 0);
    gl.glMaterialfv(GL.GL_FRONT, GL2.GL_SPECULAR, kSpecularColour, 0);
    gl.glMaterialf(GL.GL_FRONT, GL2.GL_SHININESS, kShininess);

    gl.glShadeModel(GL2.GL_FLAT);
    gl.glPolygonOffset(1.0f, 1.0f);
    gl.glEnable(GL.GL_POLYGON_OFFSET_FILL);
    addShapes(gl, false);
    gl.glDisable(GL.GL_POLYGON_OFFSET_FILL);

    gl.glDisable(GL2.GL_LIGHTING);
    gl.glEnable(GL.GL_LINE_SMOOTH);
    gl.glEnable(GL.GL_BLEND);
    gl.glColor4fv(kLineColour, 0);
    addShapes(gl, true);
    gl.glDisable(GL.GL_BLEND);
    gl.glEnable(GL2.GL_LIGHTING);
    
    gl.glPopMatrix();
    gl.glEndList();

    assert( kDisplayList > 0 );
  
  } // init()
  
  // constructor
  public UFO(float xPos, float yPos, float zPos) {
  
    mXPos = xPos;
    mYPos = yPos;
    mZPos = zPos;

    mXVel = +kMaxSpeed;
    mYVel = 0.0f;
    mZVel = 0.0f;

    mFlockId    = Env.randomInt(kNumFlocks);
    mSwitchTime = Env.randomFloat(kSwitchMinDelay, kSwitchMaxDelay);
    
    mAngle = Env.randomFloat(0.0f, 360.0f);
    mAngSpeedTarget = (Env.randomBoolean() ? +1 : -1)
                      * Env.randomFloat(kAngSpeedMin, kAngSpeedMax);
    mAngSpeed = mAngSpeedTarget;
    mAngSpeedDelay = Env.randomFloat(kAngDelayMin, kAngDelayMax);
    
    updateDirection();
    
  } // constructor
  
  // accessors
  public float getXPos() { return mXPos; }
  public float getYPos() { return mYPos; }
  public float getZPos() { return mZPos; }
  public float getXVel() { return mXVel; }
  public float getYVel() { return mYVel; }
  public float getZVel() { return mZVel; }
  
  // radii of spheres 
  public float getCollisionRadius() { return kRadius; }
  public float getVisibleRadius() { return kRadius; }
  public float getTargetRadius() { return 2.0f*kRadius; }

  // update direction details
  private void updateDirection() {

    if ( mDirection == null ) mDirection = new Quaternion();

    double vec[] = { 0.0, 0.0, -1.0 };
    mDirection.rotate(vec);
    
    double vx = mXVel,
           vy = mYVel,
           vz = mZVel;
    double d  = Math.sqrt(vx*vx + vy*vy + vz*vz);
    if ( d < 1.0e-4 ) return;
    vx /= d;
    vy /= d;
    vz /= d;
    
    double dot   = vx*vec[0] + vy*vec[1] + vz*vec[2],
           theta = Math.acos(dot);
    double ax    = vy*vec[2] - vz*vec[1],
           ay    = vz*vec[0] - vx*vec[2],
           az    = vx*vec[1] - vy*vec[0];
    Quaternion q = new Quaternion(-theta, ax, ay, az);
    mDirection = q.multiply(mDirection);

    mDirection.normalize();
    
  } // updateDirection()
  
  // change position, etc.
  public void advance(float dt, LinkedList<Sprite> sprites) {

    // change allegiance
    mSwitchTime -= dt;
    if ( mSwitchTime <= 0.0f ) {
      mFlockId = Env.randomInt(kNumFlocks);
      mSwitchTime = Env.randomFloat(kSwitchMinDelay, kSwitchMaxDelay);
    }
    
    // collect forces
    float xAcc = 0.0f, 
          yAcc = 0.0f,
          zAcc = 0.0f;
    
    // cage walls
    Cage cage = null;
    for ( Sprite sp : sprites ) {
      if ( sp instanceof Cage ) {
        cage = (Cage)sp;
        break;
      }
    }
    assert( cage != null );

    for ( Plane p : cage.getPlanes() ) {
      final float d = p.distance(mXPos, mYPos, mZPos);
      final float acc = repulsion(Math.max(0.0f, d), 
                                  kCageRange, kCageHardness);
      if ( acc > 0 ) {
        float n[] = p.normal();
        xAcc += acc*n[0];
        yAcc += acc*n[1];
        zAcc += acc*n[2];
      }
    }
    
    /*
    xAcc -= repulsion(Math.max(0.0f, Env.worldRadius() - mXPos), 
                      kCageRange, kCageHardness);
    xAcc += repulsion(Math.max(0.0f, Env.worldRadius() + mXPos), 
                      kCageRange, kCageHardness);
    yAcc -= repulsion(Math.max(0.0f, Env.worldRadius() - mYPos), 
                      kCageRange, kCageHardness);
    yAcc += repulsion(Math.max(0.0f, Env.worldRadius() + mYPos), 
                      kCageRange, kCageHardness);
    zAcc -= repulsion(Math.max(0.0f, Env.worldRadius() - mZPos), 
                      kCageRange, kCageHardness);
    zAcc += repulsion(Math.max(0.0f, Env.worldRadius() + mZPos), 
                      kCageRange, kCageHardness);
    */
    
    // collect flock details
    float xUfo   = 0.0f,
          yUfo   = 0.0f,
          zUfo   = 0.0f;
    float xv     = 0.0f,
          yv     = 0.0f,
          zv     = 0.0f;
    int   numUfo = 0,
          numVel = 0;
    
    // other obstacles
    for ( Sprite sp : sprites ) {
      if ( sp == this ) continue;
      
      float dx = mXPos - sp.getXPos(),
            dy = mYPos - sp.getYPos(),
            dz = mZPos - sp.getZPos();
      float d2 = dx*dx + dy*dy + dz*dz;
      
      float range    = 0.0f,
            hardness = 0.0f;
      if ( sp instanceof Planet ) {
        range    = kPlanetRange;
        hardness = kPlanetHardness;
      } else if ( sp instanceof Player ) {
        range    = kPlayerRange;
        hardness = kPlayerHardness;
      } else if ( sp instanceof Station ) {
        range    = kStationRange;
        hardness = kStationHardness;
      } else if ( sp instanceof WarpGate ) {
        range    = kWarpGateRange;
        hardness = kWarpGateHardness;
      } else if ( sp instanceof UFO ) {
        UFO ufo = (UFO)sp;
        range    = kUfoRange;
        hardness = kUfoHardness;
        if ( mFlockId == ufo.mFlockId ) {
          xUfo += ufo.getXPos();
          yUfo += ufo.getYPos();
          zUfo += ufo.getZPos();
          numUfo += 1;
          if ( d2 < kFlockRange*kFlockRange ) {
            xv   += ufo.getXVel();
            yv   += ufo.getYVel();
            zv   += ufo.getZVel();
            numVel += 1;
          }
        }
      } else {
        continue;
      }
      
      float f = repulsion(d2, sp.getCollisionRadius(), range, hardness);
      if ( f > 0.0f && d2 > 1.0e-6f ) {
        float d = (float)Math.sqrt(d2);
        xAcc += f*dx/d;
        yAcc += f*dy/d;
        zAcc += f*dz/d;
      }
    }

    // pull towards mean herd position
    if ( numUfo > 0 ) {
      xUfo /= numUfo; 
      yUfo /= numUfo; 
      zUfo /= numUfo;
      float dx = xUfo - mXPos,
            dy = yUfo - mYPos,
            dz = zUfo - mZPos;
      float d = (float)Math.sqrt(dx*dx + dy*dy + dz*dz);
      if ( d > 1.0f ) {
        xAcc += kHerdAccel*dx/d;
        yAcc += kHerdAccel*dy/d;
        zAcc += kHerdAccel*dz/d;
      }
    }
    
    // pull towards mean flock direction
    if ( numVel > 0 ) {
      xv /= numVel;
      yv /= numVel;
      zv /= numVel;
      float v = (float)Math.sqrt(xv*xv + yv*yv + zv*zv);
      if ( v > 0.01f ) {
        xv *= 0.9f*kMaxSpeed/v;
        yv *= 0.9f*kMaxSpeed/v;
        zv *= 0.9f*kMaxSpeed/v;
        float dx = xv - mXVel,
              dy = yv - mYVel,
              dz = zv - mZVel;
        float d = (float)Math.sqrt(dx*dx + dy*dy + dz*dz);
        if ( d > 0.01f ) {
          xAcc += kFlockAccel*dx/d;
          yAcc += kFlockAccel*dy/d;
          zAcc += kFlockAccel*dz/d;
        }
      }
    }

    // limit acceleration
    float a2 = xAcc*xAcc + yAcc*yAcc + zAcc*zAcc;
    if ( a2 > kMaxAccel*kMaxAccel ) {
      float a = (float)Math.sqrt(a2);
      xAcc *= kMaxAccel/a;
      yAcc *= kMaxAccel/a;
      zAcc *= kMaxAccel/a;
    }
    
    // accelerate
    mXVel += dt*xAcc; 
    mYVel += dt*yAcc; 
    mZVel += dt*zAcc; 
    float v2 = mXVel*mXVel + mYVel*mYVel + mZVel*mZVel;
    if ( v2 > kMaxSpeed*kMaxSpeed ) {
      float v = (float)Math.sqrt(v2);
      mXVel *= kMaxSpeed/v;
      mYVel *= kMaxSpeed/v;
      mZVel *= kMaxSpeed/v;
    }
    updateDirection();

    // move
    mXPos += dt*mXVel;
    mYPos += dt*mYVel;
    mZPos += dt*mZVel;

    // rotation
    mAngle += dt*mAngSpeed;
    if ( mAngSpeed < mAngSpeedTarget ) {
      mAngSpeed = Math.min(mAngSpeedTarget, mAngSpeed+dt*kAngSpeedDelta);
    } else if ( mAngSpeed > mAngSpeedTarget ) {
      mAngSpeed = Math.max(mAngSpeedTarget, mAngSpeed-dt*kAngSpeedDelta);
    }
    mAngSpeedDelay -= dt;
    if ( mAngSpeedDelay <= 0.0f ) {
      mAngSpeedTarget = (Env.randomBoolean() ? +1 : -1)
                        * Env.randomFloat(kAngSpeedMin, kAngSpeedMax);
      mAngSpeedDelay = Env.randomFloat(kAngDelayMin, kAngDelayMax);
    }
    
  } // Sprite.advance()
  
  // repulsive force from a barrier (linear)
  private float repulsion(float distance, float forceRange, float hardForce) {
    
    assert( distance >= 0.0f );
    assert( forceRange > 0.0f );
    assert( hardForce >= 0.0f && hardForce <= 1.0f );

    float d = (distance - hardForce*kStoppingDistance);
    float f = 1.0f - d/forceRange;
    
    if      ( f <= 0.0f ) return 0.0f;
    else if ( f >= 1.0f ) return kMaxAccel;
    else                  return f*kMaxAccel;
    
  } // repulsion()
  
  // repulsive force from a barrier (radial)
  private float repulsion(float dist2, float radius,
                          float forceRange, float hardForce) {
    
    assert( dist2 >= 0.0f );
    assert( radius >= 0.0f );
    assert( forceRange > 0.0f );
    assert( hardForce >= 0.0f && hardForce <= 1.0f );

    float r = radius + hardForce*kStoppingDistance + forceRange;
    if ( dist2 > r*r ) return 0.0f;
    
    float d = ((float)Math.sqrt(dist2) - hardForce*kStoppingDistance - radius);
    if ( d < 0.0f ) return kMaxAccel;
   
    float f = 1.0f - d/forceRange;
    assert( f >= 0.0f && f <= 1.0f );
    return f*kMaxAccel;
    
  } // repulsion()
  
  // draw the object
  public void display(GL2 gl) {
    
    assert( kDisplayList > 0 );

    double angle  = mDirection.getAngle()*180.0/Math.PI;
    double axis[] = new double[3];
    mDirection.getAxis(axis);
    
    gl.glPushMatrix();
    gl.glTranslatef(mXPos, mYPos, mZPos);
    gl.glRotatef((float)angle, (float)axis[0], (float)axis[1], (float)axis[2]);
    gl.glRotatef(mAngle, 0.0f, 0.0f, 1.0f);
    gl.glCallList(kDisplayList);
    gl.glPopMatrix();
    
  } // Sprite.display()

  // triangles and quads that make up the object
  static private void addShapes(GL2 gl, boolean outline) {

    addTriangle(gl,   0.0f,   0.0f, kMinZ, 
                    -kMaxX,   0.0f,   0.0f,
                      0.0f, +kMaxY,   0.0f, outline);
    addTriangle(gl,   0.0f,   0.0f, kMinZ, 
                      0.0f, +kMaxY,   0.0f,
                    +kMaxX,   0.0f,   0.0f, outline);
    addTriangle(gl,   0.0f,   0.0f, kMinZ, 
                    +kMaxX,   0.0f,   0.0f,
                      0.0f, -kMaxY,   0.0f, outline);
    addTriangle(gl,   0.0f,   0.0f, kMinZ, 
                      0.0f, -kMaxY,   0.0f,
                    -kMaxX,   0.0f,   0.0f, outline);
    
    addTriangle(gl,   0.0f,   0.0f, kMaxZ, 
                    +kMaxX,   0.0f,   0.0f,
                      0.0f, +kMaxY,   0.0f, outline);
    addTriangle(gl,   0.0f,   0.0f, kMaxZ, 
                      0.0f, +kMaxY,   0.0f,
                    -kMaxX,   0.0f,   0.0f, outline);
    addTriangle(gl,   0.0f,   0.0f, kMaxZ, 
                    -kMaxX,   0.0f,   0.0f,
                      0.0f, -kMaxY,   0.0f, outline);
    addTriangle(gl,   0.0f,   0.0f, kMaxZ, 
                      0.0f, -kMaxY,   0.0f,
                    +kMaxX,   0.0f,   0.0f, outline);
    
  } // addShapes()
  
} // class UFO
