/*
 *  WarpGate.java
 *  Copyright Simon Hern 2009
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package Space;

import java.util.LinkedList;
import javax.media.opengl.*;

// a stationary spinning object
public class WarpGate extends Sprite {

  // material details
  private static final float kDiffuseColour[]  = { 0.5f, 0.5f, 0.5f, 1.0f },
                             kSpecularColour[] = { 0.8f, 0.8f, 0.8f, 1.0f };
  private static final float kShininess        = 8.0f;
  
  // outline details
  private static final float kLineColour[] = { 0.2f, 0.2f, 0.2f, 1.0f };
  
  // object shape
  private static final float kRingRadius  = 400.0f,
                             kRingWidth   = 0.15f*kRingRadius;
  private static final float kHeartRadius = 120.0f;
  
  // motion
  private static final float kAngSpeed[] = { 299.0f, 19.0f, 
                                             10.0f, 
                                             51.0f, 83.0f, 166.0f };
  
  // bounding sphere
  private static final float kRadius = (float)Math.sqrt(
                                                kRingRadius*kRingRadius
                                              + kRingWidth*kRingWidth );
  private static final float kAtmosphere = 50.0f;
  
  // display list
  private static int kDisplayList1 = 0,
                     kDisplayList2 = 0;

  // position (of centre of object)
  private float mXPos,
                mYPos,
                mZPos;
  
  // rotation
  private float mAngle[];
  
  // prepare display list
  static public void init(GL2 gl) {
  
    if ( kDisplayList1 != 0 ) return; // already initialized

    // outer gate
    kDisplayList1 = gl.glGenLists(1);
    gl.glNewList(kDisplayList1, GL2.GL_COMPILE);
    gl.glPushMatrix();
    
    gl.glMaterialfv(GL.GL_FRONT, GL2.GL_AMBIENT_AND_DIFFUSE, kDiffuseColour, 0);
    gl.glMaterialfv(GL.GL_FRONT, GL2.GL_SPECULAR, kSpecularColour, 0);
    gl.glMaterialf(GL.GL_FRONT, GL2.GL_SHININESS, kShininess);

    gl.glShadeModel(GL2.GL_FLAT);
    gl.glPolygonOffset(1.0f, 1.0f);
    gl.glEnable(GL.GL_POLYGON_OFFSET_FILL);
    addRingShapes(gl, false);
    gl.glDisable(GL.GL_POLYGON_OFFSET_FILL);

    gl.glDisable(GL2.GL_LIGHTING);
    gl.glEnable(GL.GL_LINE_SMOOTH);
    gl.glEnable(GL.GL_BLEND);
    gl.glColor4fv(kLineColour, 0);
    addRingShapes(gl, true);
    gl.glDisable(GL.GL_BLEND);
    gl.glEnable(GL2.GL_LIGHTING);
    
    gl.glPopMatrix();
    gl.glEndList();

    // inner gate
    kDisplayList2 = gl.glGenLists(1);
    gl.glNewList(kDisplayList2, GL2.GL_COMPILE);
    gl.glPushMatrix();
    
    gl.glMaterialfv(GL.GL_FRONT, GL2.GL_AMBIENT_AND_DIFFUSE, kDiffuseColour, 0);
    gl.glMaterialfv(GL.GL_FRONT, GL2.GL_SPECULAR, kDiffuseColour, 0);
    gl.glMaterialf(GL.GL_FRONT, GL2.GL_SHININESS, kShininess);

    gl.glShadeModel(GL2.GL_FLAT);
    gl.glPolygonOffset(1.0f, 1.0f);
    gl.glEnable(GL.GL_POLYGON_OFFSET_FILL);
    addHeartShapes(gl, false);
    gl.glDisable(GL.GL_POLYGON_OFFSET_FILL);

    gl.glDisable(GL2.GL_LIGHTING);
    gl.glEnable(GL.GL_LINE_SMOOTH);
    gl.glEnable(GL.GL_BLEND);
    gl.glColor4fv(kLineColour, 0);
    addHeartShapes(gl, true);
    gl.glDisable(GL.GL_BLEND);
    gl.glEnable(GL2.GL_LIGHTING);
    
    gl.glPopMatrix();
    gl.glEndList();

  } // init()
  
  // constructor (angles in degrees)
  public WarpGate(float xPos, float yPos, float zPos) {

    mXPos = xPos;
    mYPos = yPos;
    mZPos = zPos;
    
    mAngle = new float[kAngSpeed.length];
    for ( int k = 0 ; k < mAngle.length ; k++ ) {
      mAngle[k] = 90.0f;
    }
    
  } // constructor
  
  // accessors
  public float getXPos() { return mXPos; }
  public float getYPos() { return mYPos; }
  public float getZPos() { return mZPos; }
  
  // radii of spheres
  public float getCollisionRadius() { return (kRadius + kAtmosphere); }
  public float getVisibleRadius()   { return kRadius; }
  public float getTargetRadius()    { return 0.4f*kRadius; }
  
  // change position, etc.
  public void advance(float dt, LinkedList<Sprite> sprites) {

    for ( int k = 0 ; k < mAngle.length ; k++ ) {
      mAngle[k] += dt*kAngSpeed[k];
    }
    
  } // Sprite.advance()
  
  // draw the object
  public void display(GL2 gl) {
    
    assert( kDisplayList1 > 0 );

    float s1 = kRingRadius*(float)Math.cos(Math.PI/8),
          s2 = (float)Math.sqrt(kRingRadius*kRingRadius + kRingWidth*kRingWidth);
    float shrink = 0.99f*s1/s2;
    
    gl.glDisable(GL.GL_CULL_FACE);
    
    gl.glPushMatrix();
    gl.glTranslatef(mXPos, mYPos, mZPos);
    gl.glRotatef(mAngle[2], 1.0f, 0.0f, 0.0f);
    for ( int k = 3 ; k < mAngle.length ; k++ ) {
      float z = (k%2==0) ? 0.0f : 1.0f,
            x = 1.0f - z;
      gl.glRotatef(mAngle[k], x, 0.0f, z);
      gl.glCallList(kDisplayList1);
      gl.glScalef(shrink, shrink, shrink);
    }
    gl.glPopMatrix();
    
    gl.glEnable(GL.GL_CULL_FACE);

    gl.glPushMatrix();
    gl.glTranslatef(mXPos, mYPos, mZPos);
    gl.glRotatef(mAngle[0], 0.0f, 0.0f, 1.0f);
    gl.glRotatef(mAngle[0], 1.0f, 0.0f, 0.0f);
    gl.glCallList(kDisplayList2);
    gl.glPopMatrix();
    
  } // Sprite.display()

  // triangles and quads that make up the outer gate
  static private void addRingShapes(GL2 gl, boolean outline) {

    final int numSides = 8;
    float dTheta = 2.0f*(float)Math.PI/numSides;
    for ( int k = 0 ; k < numSides ; k++ ) {
      float theta1 = (k-0.5f)*dTheta,
            theta2 = ( (k == numSides-1) ? -0.5f : (k+0.5f) )*dTheta;
      float x1 = kRingRadius*(float)Math.cos(theta1),
            z1 = kRingRadius*(float)Math.sin(theta1),
            x2 = kRingRadius*(float)Math.cos(theta2),
            z2 = kRingRadius*(float)Math.sin(theta2);
      addQuad(gl, x1, +kRingWidth, z1,
                  x2, +kRingWidth, z2,
                  x2, -kRingWidth, z2,
                  x1, -kRingWidth, z1, outline);
    }

  } // addRingShapes()
  
  // triangles and quads that make up the inner gate
  static private void addHeartShapes(GL2 gl, boolean outline) {

    float x1 = 0.0f,
          y1 = kHeartRadius,
          z1 = 0.0f;
    
    double theta1 = 2.0*Math.asin(Math.sqrt(2.0/3.0));
    float x2 = 0.0f,
          y2 = kHeartRadius*(float)Math.cos(theta1),
          z2 = kHeartRadius*(float)Math.sin(theta1);
         
    double theta2 = 2.0*Math.PI/3.0;
    float x3 = z2*(float)Math.sin(theta2),
          y3 = y2,
          z3 = z2*(float)Math.cos(theta2);
    float x4 = z2*(float)Math.sin(2*theta2),
          y4 = y2,
          z4 = z2*(float)Math.cos(2*theta2);
    
    addTriangle(gl, x1, y1, z1,
                    x2, y2, z2,
                    x3, y3, z3, outline);
    addTriangle(gl, x1, y1, z1,
                    x3, y3, z3,
                    x4, y4, z4, outline);
    addTriangle(gl, x1, y1, z1,
                    x4, y4, z4,
                    x2, y2, z2, outline);
    addTriangle(gl, x2, y2, z2,
                    x4, y4, z4,
                    x3, y3, z3, outline);

  } // addHeartShapes()
  
} // class Station
