/**
 *  Voronoi3D.java
 *  Copyright Simon Hern 2009
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package Space.patterns;

import java.util.ArrayList;
import java.util.Random;

// Voronoi/Worley-style cellular function defined in three-dimensions
// (the function is periodic along all axes with unit period)
public class Voronoi3D extends Function {

  // order in which to search the 26 neighbouring cubes
  // (search adjoining faces, then edges, then corners - the idea is to try
  // the most likely candidates first and terminate the search quickly)
  static private final int kNeighbourOrder[][] 
         = { {+1,0,0}, {-1,0,0}, {0,+1,0}, {0,-1,0}, {0,0,+1}, {0,0,-1},
             {+1,+1,0}, {+1,-1,0}, {-1,+1,0}, {-1,-1,0}, 
             {+1,0,+1}, {+1,0,-1}, {-1,0,+1}, {-1,0,-1}, 
             {0,+1,+1}, {0,+1,-1}, {0,-1,+1}, {0,-1,-1},
             {+1,+1,+1}, {+1,+1,-1}, {+1,-1,+1}, {+1,-1,-1},
             {-1,+1,+1}, {-1,+1,-1}, {-1,-1,+1}, {-1,-1,-1} };
    
  // xyz-coordinates of random points in the cube [0,1]x[0,1]x[0,1]
  private ArrayList<float[]> mPoints;
  
  // tree structure for spatial search
  private VoronoiTree mSearchTree;
  
  // constructor
  public Voronoi3D(long seed, int numPoints) {
    
    Random rand = new Random(seed);
    
    assert( numPoints >= 2 );
    mPoints = new ArrayList<float[]>(numPoints);
    
    for ( int k = 0 ; k < numPoints ; k++ ) {
      float xyz[] = { rand.nextFloat(), 
                      rand.nextFloat(), 
                      rand.nextFloat() };
      mPoints.add(xyz);
    }
    
    mSearchTree = new VoronoiTree(mPoints);
    
  } // constructor
  
  // returned value is the distance-squared to the nearest point
  // (use the search tree if defined, otherwise brute force)
  @Override
  public float value(float x, float y, float z) {
    
    if ( mSearchTree != null ) return valueFromTree(x, y, z);
    else                       return valueBruteForce(x, y, z);
    
  } // Function.value()
      
  // use the search tree to find the nearest point
  // (periodicity means we may have to use the search tree several times)
  private float valueFromTree(float x, float y, float z) {
    
    assert( mSearchTree != null );
    
    // first find the nearest point in the cube [0,1]x[0,1]x[0,1]
    
    float pos[] = { x - (float)Math.floor(x), 
                    y - (float)Math.floor(y), 
                    z - (float)Math.floor(z) };
    float d2Min = mSearchTree.search(pos);
    
    // now check for nearer points in the 26 neighbouring cubes
    
    float delta0[]   = new float[3],
          delta1[]   = new float[3],
          posNeigh[] = new float[3];
    for ( int i = 0 ; i < 3 ; i++ ) {
      final float d = pos[i];
      delta0[i] = d*d;
      delta1[i] = (1.0f-d)*(1.0f-d);
    }
    for ( int neigh[] : kNeighbourOrder ) {
      assert( neigh != null && neigh.length == 3 );
      float d2Neigh = 0.0f;
      for ( int i = 0 ; i < 3 ; i++ ) {
        assert( neigh[i] >= -1 && neigh[i] <= +1 );
        if      ( neigh[i] > 0 ) d2Neigh += delta0[i];
        else if ( neigh[i] < 0 ) d2Neigh += delta1[i];
      }
      if ( d2Neigh <= d2Min ) {
        for ( int i = 0 ; i < 3 ; i++ ) posNeigh[i] = pos[i] + neigh[i];
        d2Min = mSearchTree.search(posNeigh, d2Min);
      }
    }
    
    assert( Math.abs(d2Min - valueBruteForce(x,y,z)) < 1.0e-6f );
      
    return d2Min;
      
  } // valueFromTree()
      
  // test all points to find the nearest
  private float valueBruteForce(float x, float y, float z) {
    
    float d2Min = 1.0e6f;
    for ( int k = 0 ; k < mPoints.size() ; k++ ) {
      float point[] = mPoints.get(k);
      float dx = fold(x - point[0]),
            dy = fold(y - point[1]),
            dz = fold(z - point[2]);
      float d2 = dx*dx + dy*dy + dz*dz;
      if ( d2 < d2Min ) {
        d2Min = d2;
      }
    }    
    return d2Min;
   
  } // valueBruteForce()

  // fold into the range [-0.5,+0.5]
  private float fold(float x) {

    float d = (float)Math.floor(x + 0.5f);
    return (x - d);
    
  } // fold()
  
} // class Voronoi3D
