/**
 *  VoronoiSphere.java
 *  Copyright Simon Hern 2009
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package Space.patterns;

import java.util.ArrayList;
import java.util.Random;

// Voronoi/Worley-style cellular function defined on the unit sphere
public class VoronoiSphere extends Function {

  // xyz-coordinates of random points on the unit sphere
  private ArrayList<float[]> mPoints;
  
  // tree structure for spatial search
  private VoronoiTree mSearchTree;
  
  // constructor
  public VoronoiSphere(long seed, int numPoints) {
    
    Random rand = new Random(seed);
    
    assert( numPoints >= 2 );
    mPoints = new ArrayList<float[]>(numPoints);
    
    while ( mPoints.size() < numPoints ) {
      float x = 2.0f*rand.nextFloat() - 1.0f,
            y = 2.0f*rand.nextFloat() - 1.0f,
            z = 2.0f*rand.nextFloat() - 1.0f;
      float d2 = x*x + y*y + z*z;
      if ( d2 <= 1.0f && d2 > 0.01 ) {
        float d = (float)Math.sqrt(d2);
        float xyz[] = { x/d, y/d, z/d };
        mPoints.add(xyz);
      }
    }
    
    mSearchTree = new VoronoiTree(mPoints);
    
  } // constructor
  
  // returned value is the distance-squared to the nearest point
  // (use the search tree if defined, otherwise brute force)
  @Override
  public float value(float x, float y, float z) {
    
    if ( mSearchTree != null ) {

      float pos[] = {x, y, z};
      return mSearchTree.search(pos);
      
    } else {
    
      float d2Min = 1.0e6f;
      for ( int k = 0 ; k < mPoints.size() ; k++ ) {
        float point[] = mPoints.get(k);
        float dx = x - point[0],
              dy = y - point[1],
              dz = z - point[2];
        float d2 = dx*dx + dy*dy + dz*dz;
        if ( d2 < d2Min ) {
          d2Min = d2;
        }
      }    
      return d2Min;
      
    }
    
  } // Function.value()

} // class VoronoiSphere
