/**
 *  VoronoiTree.java
 *  Copyright Simon Hern 2009
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package Space.patterns;

import java.util.*;

// Tree structure for speeding up Voronoi searches (finding the distance
// to the nearest point in a list).
//
// The points are arranged into a kd-tree.  The algorithm is a standard
// nearest-neighbour search.  (The search starts by finding the appropriate 
// leaf node given the spatial subdivision, then climbs back to the root 
// node while checking in case any of the other branches contain a nearer 
// point.)
// 
// The algorithm returns the distance-squared to nearest point.  The class
// will work with points in any number of spatial dimensions.

public class VoronoiTree {

  // a leaf or branch in the tree
  private abstract class Node {
    abstract float search(float position[], float minDist2);
  } // class Node
  
  // a branch splits into two sub-trees at a point along one of the axes
  private class Branch extends Node {
    private int mAxis;
    private float mValue;
    private Node mBelow, mAbove;
    public Branch(int axis, float value, Node below, Node above) {
      mAxis = axis;
      mValue = value;
      mBelow = below;
      mAbove = above;
    }
    public float search(float position[], float minDist2) {
      final float dv = position[mAxis] - mValue;
      if ( dv > 0.0f ) {
        minDist2 = mAbove.search(position, minDist2);
        if ( dv*dv < minDist2 ) minDist2 = mBelow.search(position, minDist2);
      } else {
        minDist2 = mBelow.search(position, minDist2);
        if ( dv*dv < minDist2 ) minDist2 = mAbove.search(position, minDist2);        
      }
      return minDist2;
    }
  } // class Branch
  
  // a leaf is a single point
  private class Leaf extends Node {
    private float mPosition[];
    public Leaf(float pos[]) { 
      mPosition = pos; 
    }
    public float search(float position[], float minDist2) {
      assert( position.length == mPosition.length );
      float d2 = 0.0f;
      for ( int k = 0 ; k < mPosition.length ; k++ ) {
        final float d = position[k] - mPosition[k];
        d2 += d*d;
      }
      return Math.min(minDist2, d2);
    }
  } // class Leaf
  
  // root node of the tree
  private Node mRoot;
  
  // dimension of the Voronoi points (all float arrays must have this length)
  private int mDimension;
  
  // constructor
  public VoronoiTree(ArrayList<float[]> points) {
    
    assert( points != null );
    mDimension = 0;
    mRoot = buildTree(points);
    
  } // constructor
  
  // construct a search tree (or part of one) out of a list of points
  private Node buildTree(ArrayList<float[]> points) {
    
    assert( points != null && points.size() > 0 );

    float firstPoint[] = points.get(0);
    assert( firstPoint != null );
    
    if ( mDimension == 0 ) mDimension = firstPoint.length;

    // a single point becomes a tree leaf
    
    if ( points.size() == 1 ) {
      return new Leaf(firstPoint);
    }
    
    // derive the bounding box of the points 
    
    float min[] = firstPoint.clone(),
          max[] = firstPoint.clone();
    for ( float point[] : points ) {
      assert( point != null && point.length == mDimension );
      for ( int k = 0 ; k < mDimension ; k++ ) {
        min[k] = Math.min(min[k], point[k]);
        max[k] = Math.max(max[k], point[k]);
      }
    }

    // split the bounding box at the mid-point of the longest axis
    
    int splitAxis = 0;
    for ( int k = 1 ; k < mDimension ; k++ ) {
      if ( max[k] - min[k] > max[splitAxis] - min[splitAxis] ) {
        splitAxis = k;
      }
    }
    final float splitValue = 0.5f*(max[splitAxis] + min[splitAxis]);

    // split the points into two lists
    
    ArrayList<float[]> below = new ArrayList<float[]>(points.size()),
                       above = new ArrayList<float[]>(points.size());
    for ( float[] point : points ) {
      if ( point[splitAxis] > splitValue ) above.add(point);
      else                                 below.add(point);
    }
    
    // special case: points are identical
    
    if ( below.size() == 0 || above.size() == 0 ) {
      return new Leaf(firstPoint);
    }
    
    // add a branch to the tree
    
    return new Branch(splitAxis, splitValue,
                      buildTree(below), buildTree(above));
    
  } // buildTree()
  
  // returns the distance-squared to the nearest point in the tree
  public float search(float position[]) {
    
    assert( position != null && position.length == mDimension );
    return mRoot.search(position, 1.0e8f);
    
  } // search()
  
  // returns the distance-squared to the nearest point in the tree
  // or minDist2 if no nearer points are found  
  public float search(float position[], float minDist2) {
    
    assert( position != null && position.length == mDimension );
    return mRoot.search(position, minDist2);
    
  } // search()
  
} // class VoronoiTree
