/**
 *  Globe.java
 *  Copyright Simon Hern 2009
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package Space.planets;

import Space.Env;
import Space.patterns.Col24;
import java.awt.image.*;
import javax.media.opengl.*;
import com.sun.opengl.util.texture.*;
import com.sun.opengl.util.texture.awt.AWTTextureIO;

// a textured sphere that represents a planet
public class Globe {

  // planet radius (pixels) at which subdivision depth changes
  // (these values have been plucked from thin air, but seem to work okay)
  static private final int kDepthTransitions[] = { 2, 4, 6, 20, 200, 600 }; 
  
  // slight reluctance for the subdivision depth to decrease
  static private final float kRadiusSafetyFactor = 1.02f;
  
  // a Halo object (atmospheric effect) is shared between Globes
  static private Halo kHalo;
  
  // function for generating the texture for the sphere surface
  private SphericalSkin mSkin;
  
  // size of texture to use for the surface (should be power of 2)
  private int mTextureSize;
  
  // texture for the sphere surface
  private Texture mTexture;
  
  // overall surface colour (4-element RGBA array)
  private float[] mAverageColour;
  
  // subdivision depth of the sphere
  private int mDepth;
  
  // constructor (texture size should be a power of two)
  public Globe(SphericalSkin skin, int textureSize) {
  
    assert( skin != null );
    mSkin = skin;
    
    assert( textureSize > 0 );
    mTextureSize = textureSize;
    
    mTexture = null;
    mAverageColour = null;
    mDepth = 5;
    
  } // constructor

  // build the texture for the sphere surface, and prepare other resources 
  public void prepare(GL2 gl) {

    if ( mTexture != null ) return;

    mAverageColour = new float[]{ 0.0f, 0.0f, 0.0f, 1.0f };

    long nanos = System.nanoTime();
    
    BufferedImage image = new BufferedImage(mTextureSize, mTextureSize, 
                                            BufferedImage.TYPE_INT_RGB);
    WritableRaster raster = image.getRaster();
    int pixels[] = ((DataBufferInt)(raster.getDataBuffer())).getData();
    
    float xyz[] = new float[3];
    Sphere sphere = Sphere.getInstance();
    int pixInd = 0;
    for ( int iy = 0 ; iy < mTextureSize ; iy++ ) {
      for ( int ix = 0 ; ix < mTextureSize ; ix++ ) {
        sphere.texelToXyz(ix, iy, mTextureSize, xyz);          
        int rgb = mSkin.colour(xyz[0], xyz[1], xyz[2]);
        pixels[pixInd++] = rgb;

        mAverageColour[0] += Col24.red(rgb);
        mAverageColour[1] += Col24.green(rgb);
        mAverageColour[2] += Col24.blue(rgb);
      }
    }
    mTexture = AWTTextureIO.newTexture(image, true);
    
    if ( Env.debugMode() ) {
      nanos = System.nanoTime() - nanos;
      System.out.println("Generated planet texture (" 
                         + String.format("%.2f", (nanos*1.0e-9f))
                         + " sec)");
    }

    final int num = mTextureSize*mTextureSize;
    for ( int k = 0 ; k < 3 ; k++ ) mAverageColour[k] /= 255.0f*num;

    if ( kHalo == null ) {
      kHalo = new Halo();
      kHalo.prepare(gl);
    }
    
  } // prepare()

  // discard resources
  public void dispose(GL2 gl) {
    
    mTexture.destroy(gl);
    mTexture = null;
    
  } // dispose()
  
  // mean colour of surface (RGBA array)
  public float[] averageColour() { return mAverageColour; }
  
  // the maximum subdivision depth for this shape
  public int maxDepth() { return Sphere.getInstance().maxDepth(); }
  
  // specify the sphere's subdivision depth explicitly
  public void setDepth(int depth) { 
    
    assert( depth >= 0 && depth <= maxDepth() );
    mDepth = depth;
    
  } // setDepth()
  
  // decide the sphere's subdivision depth based on its size on the screen
  public void setDepthFromSize(float radiusPixels) {
    
    assert( radiusPixels >= 0 );
    
    int suggestedDepth = kDepthTransitions.length;
    for ( int k = 0 ; k < kDepthTransitions.length ; k++ ) {
      if ( kDepthTransitions[k] > radiusPixels ) {
        suggestedDepth = k;
        break;
      }
    }
    assert( suggestedDepth <= maxDepth() );
    
    if ( suggestedDepth > mDepth ) {
      //System.out.println("depth: " + mDepth + " -> " + suggestedDepth);
      mDepth = suggestedDepth;
    } else if ( suggestedDepth < mDepth ) {
      final float safeRadius = radiusPixels*kRadiusSafetyFactor;
      if ( kDepthTransitions[suggestedDepth] > safeRadius ) {
        //System.out.println("depth: " + mDepth + " -> " + suggestedDepth);
        mDepth = suggestedDepth;
      }
    }
    
  } // setDepthFromSize()
  
  // draw the sphere (unit size, centred at origin)
  public void display(GL2 gl) {
    
    if ( mTexture == null ) prepare(gl);
    Sphere.getInstance().display(gl, mDepth, mTexture);
    
  } // display()
  
  // draw an atmospheric effect around the sphere
  public void displayHalo(GL2 gl, float haloScale, float[] viewXyz) {

    assert( haloScale > 1.0f );
    assert( viewXyz != null && viewXyz.length == 3 );
    kHalo.display(gl, haloScale, mAverageColour, viewXyz);

  } // displayHalo()
  
} // class Globe
