/**
 *  Halo.java
 *  Copyright Simon Hern 2009
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package Space.planets;

import java.nio.*;
import javax.media.opengl.*;

// Atmospheric glow around a sphere (typically representing a planet).
// 
// The glow is drawn for a unit sphere centred at the origin.  (The caller
// should translate and scale as necessary.)  The glow is a transparency
// effect based on a 1D alpha texture.  The texture is opaque at the
// sphere's surface and fully transparent at the effect's maximum range.
//
// What is drawn depends on the position of the viewer.  The surface of the
// effect needs to be perpendicular to both the viewing direction and the
// surface of the sphere.  For a viewer at infinity, this means that the  
// effect is drawn on a plane through the centre of the sphere.  For a 
// viewer at finite distance, the effect is drawn on a cone whose tip is at
// the centre of the sphere.
//
// The triangles making up the effect are drawn using the normals at the 
// sphere surface, so the halo is affected by lighting in the same way as 
// the sphere is.

public class Halo {

  // inner edge of halo is slightly inside the planet
  static final private float kMinRadius = 0.99f;
  
  // opacity (blend alpha) at inner edge
  static final private float kMaxAlpha = 1.0f;
  
  // no specular material colour
  static final float kSpecularColour[] = { 0.0f, 0.0f, 0.0f, 1.0f };
  
  // number of points in the blend texture
  static final private int kTextureSize = 8; 

  // texture coordinates at inner and outer edges
  static final private float kMinTexCoord = 0.5f/kTextureSize,
                             kMaxTexCoord = 1.0f - kMinTexCoord;

  // number of quads to draw to create the effect
  // (could make this number adapt to the size of the effect on screen?)
  static final private int kNumSections = 256;
  
  // OpenGL identifier for the texture
  private int mTextureName;
  
  // the texture data, an array of alpha values (0 to 255)
  private ByteBuffer mData;
  
  // constructor
  public Halo() {
    
    mTextureName = 0;
  
  } // constructor
  
  // create the 1D texture for the atmosphere effect
  public void prepare(GL2 gl) {

    if ( mData != null ) return;
    
    mData = ByteBuffer.allocateDirect(kTextureSize);
    mData.order(ByteOrder.nativeOrder());
    mData.clear();
    for ( int k = 0 ; k < kTextureSize ; k++ ) {
      final float x = k/(float)(kTextureSize-1);
      final float y = kMaxAlpha*(1-x)*(1-x);
      mData.put((byte)Math.round(255*y));
    }
    mData.flip();
    
    int names[] = new int[1];
    gl.glGenTextures(1, names, 0);
    mTextureName = names[0];
    
    gl.glBindTexture(GL2.GL_TEXTURE_1D, mTextureName);
    
    gl.glTexParameteri(GL2.GL_TEXTURE_1D, 
                       GL.GL_TEXTURE_WRAP_S, GL2.GL_CLAMP_TO_EDGE);
    gl.glTexParameteri(GL2.GL_TEXTURE_1D, 
                       GL.GL_TEXTURE_MAG_FILTER, GL.GL_LINEAR);
    gl.glTexParameteri(GL2.GL_TEXTURE_1D, 
                       GL.GL_TEXTURE_MIN_FILTER, GL.GL_LINEAR);
    
    gl.glTexImage1D(GL2.GL_TEXTURE_1D, 0, GL2.GL_ALPHA,
                    kTextureSize, 0, 
                    GL2.GL_ALPHA, GL.GL_UNSIGNED_BYTE, mData);
    
  } // prepare()
  
  // given the viewer position (or unit vector to a viewer at infinity)
  // construct orthogonal axes such that the z-axis points to the viewer
  // and return the angle for the 'cone' on which the effect is drawn
  private float defineAxes(float[] view, float[] x, float[] y, float[] z) {
    
    assert( view != null && view.length == 3 );
    assert( x != null && y.length == 3 );
    assert( y != null && y.length == 3 );
    assert( z != null && z.length == 3 );
    
    final float v2 = view[0]*view[0] + view[1]*view[1] + view[2]*view[2],
                v  = (float)Math.sqrt(v2);
    assert( v > 1.0e-6 );
    
    for ( int k = 0 ; k < 3 ; k++ ) z[k] = view[k]/v;

    int minIndex = 0;
    if ( Math.abs(z[1]) < Math.abs(z[minIndex]) ) minIndex = 1;
    if ( Math.abs(z[2]) < Math.abs(z[minIndex]) ) minIndex = 2;
    
    for ( int k = 0 ; k < 3 ; k++ ) x[k] = 0.0f;
    x[minIndex] = 1;
    
    y[0] = x[1]*z[2] - x[2]*z[1];
    y[1] = x[2]*z[0] - x[0]*z[2];
    y[2] = x[0]*z[1] - x[1]*z[0];
    
    final float y2 = y[0]*y[0] + y[1]*y[1] + y[2]*y[2],
                yy = (float)Math.sqrt(y2);
    for ( int k = 0 ; k < 3 ; k++ ) y[k] /= yy;
    
    x[0] = y[1]*z[2] - y[2]*z[1];
    x[1] = y[2]*z[0] - y[0]*z[2];
    x[2] = y[0]*z[1] - y[1]*z[0];
    
    if ( v <= 1.000001f) return 0.0f;
    
    final float angle = (float)Math.asin(1.0f/v);
    assert( angle >= 0.0f );
    return angle;
    
  } // defineAxes()
  
  // draw the halo effect
  // (if the viewer position is a unit vector then it is treated as the
  // direction to a viewer at infinite distance)
  public void display(GL2 gl, float maxRadius, 
                      float[] colour, float[] viewXyz) {

    assert( gl != null );
    assert( maxRadius > 1.0f );
    assert( colour != null && colour.length == 4 );
    assert( viewXyz != null && viewXyz.length == 3 );
    
    float xAxis[] = new float[3],
          yAxis[] = new float[3],
          zAxis[] = new float[3];
    float phi = defineAxes(viewXyz, xAxis, yAxis, zAxis);

    float cph = (float)Math.cos(phi),
          sph = (float)Math.sin(phi);
    float vec[] = new float[3];

    gl.glEnable(GL.GL_BLEND);
    gl.glDepthMask(false);
    gl.glDisable(GL.GL_CULL_FACE);
    gl.glEnable(GL2.GL_TEXTURE_1D);
    
    gl.glMaterialfv(GL.GL_FRONT, GL2.GL_AMBIENT_AND_DIFFUSE, colour, 0);
    gl.glMaterialfv(GL.GL_FRONT, GL2.GL_SPECULAR, kSpecularColour, 0);
    gl.glShadeModel(GL2.GL_SMOOTH);
    
    gl.glBindTexture(GL2.GL_TEXTURE_1D, mTextureName);
    gl.glTexEnvf(GL2.GL_TEXTURE_ENV, GL2.GL_TEXTURE_ENV_MODE, GL2.GL_REPLACE);

    gl.glBegin(GL2.GL_TRIANGLE_STRIP);
    for ( int k = 0 ; k <= kNumSections ; k++ ) {
      final int n = ( k < kNumSections ? k : 0 );
      float theta = (n/(float)kNumSections)*2.0f*(float)Math.PI;
      float cth = (float)Math.cos(theta),
            sth = (float)Math.sin(theta);
      for ( int i = 0 ; i < 3 ; i++ ) {
        vec[i] = cph*(cth*xAxis[i] + sth*yAxis[i]) + sph*zAxis[i];
      }
      
      gl.glTexCoord1f(kMinTexCoord);
      gl.glNormal3f(vec[0], vec[1], vec[2]);
      gl.glVertex3f(kMinRadius*vec[0], kMinRadius*vec[1], kMinRadius*vec[2]);
      gl.glTexCoord1f(kMaxTexCoord);
      gl.glVertex3f(maxRadius*vec[0], maxRadius*vec[1], maxRadius*vec[2]);
    }
    gl.glEnd();
    
    gl.glDisable(GL2.GL_TEXTURE_1D);
    gl.glEnable(GL.GL_CULL_FACE);
    gl.glDepthMask(true);
    gl.glDisable(GL.GL_BLEND);
    
  } // display()
  
} // class Halo
