/**
 *  Octant.java
 *  Copyright Simon Hern 2009
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package Space.planets;

import Space.Env;
import java.nio.*;
import java.util.*;
import javax.media.opengl.*;

// Approximation of the unit sphere restricted to x >= 0, y >= 0, z >= 0.
// A complete unit sphere can be made by repeating the octant shape eight
// times with different rotations.
// 
// The approximation is built up by subdivision, starting from one face of
// an octahedron.  Each stage of subdivision increases the number of faces
// by a factor of four.  New vertices are added at the midpoints of existing 
// edges, projected onto the unit sphere.
//
// The vertices are labelled (i,j), where 0 <= i < N, 0 <= j < N, i+j < N,
// and N depends on the maximum depth of subdivision.  Vertex (0,0) has
// xyz-coordinates (0,1,0), vertex (N-1,0) has coordinates (1,0,0), and
// vertex (0,N-1) has coordinates (0,0,1).
//
// The vertices are stored in one-dimensional buffer objects in the order
// (0,0), (1,0), (2,0), ..., (0,1), (1,1), (2,1), etc.
//
// Texture coordinates (u,v) are related to the vertex labelling (i,j) by
// u = i/(N-1), v = j/(N-1).  The octant therefore uses one triangular
// half of a square texture.  (The other half of the texture will typically
// be used by another octant.) 

public class Octant {

  // maximum depth of subdivision (0 => one face, no subdivision)
  private int mSubdivisionDepth;
  
  // total number of vertices at the maximum subdivision depth
  private int mNumVertices;
  
  // index range for (i,j) labelling of vertices
  private int mNumIVerts;
  
  // (x,y,z) coordinates of vertices at the maximum subdivision depth
  private FloatBuffer mVertexArray;
  
  // (u,v) texture coordinates at the vertices
  private FloatBuffer mTexCoordArray;
  
  // arrays of faces at different subdivision depths
  private ShortBuffer mFaceArrays[];
  
  // OpenGL buffer objects for the different arrays
  private int mVertexBufferObj,
              mTexCoordBufferObj;
  private int mFaceBufferObj[];
  
  // depth of subdivision currently bound for drawing
  private int mBoundDepth;
  
  // constructor
  public Octant(int minDepth, int maxDepth) {
    
    assert( minDepth >= 0 );
    assert( maxDepth >= minDepth );
    mSubdivisionDepth = maxDepth;

    makeVertices();
    makeTexCoords();
    
    mFaceArrays = new ShortBuffer[mSubdivisionDepth+1];
    for ( int k = minDepth ; k <= maxDepth ; k++ ) {
      mFaceArrays[k] = makeFaces(k);
    }

    mVertexBufferObj = mTexCoordBufferObj = 0;
    mFaceBufferObj = new int[mSubdivisionDepth+1];
    Arrays.fill(mFaceBufferObj, 0);
    
    mBoundDepth = -1;
    
  } // constructor

  // the maximum subdivision depth for this shape
  public int maxDepth() { return mSubdivisionDepth; }
  
  // calculate vertex coordinates by subdividing one face of an octahedron
  private void makeVertices() {
    
    mNumIVerts = (1 << mSubdivisionDepth) + 1;
    mNumVertices = mNumIVerts*(mNumIVerts+1)/2;

    mVertexArray = Env.allocateFloatBuffer(3*mNumVertices);
    
    setVertexArray(0, 0);
    mVertexArray.put(0.0f);
    mVertexArray.put(1.0f);
    mVertexArray.put(0.0f);
    
    setVertexArray(mNumIVerts-1, 0);
    mVertexArray.put(0.0f);
    mVertexArray.put(0.0f);
    mVertexArray.put(1.0f);
    
    setVertexArray(0, mNumIVerts-1);
    mVertexArray.put(1.0f);
    mVertexArray.put(0.0f);
    mVertexArray.put(0.0f);

    subdivideFace(0, 0, mNumIVerts-1);
    
  } // makeVertices

  // subdivide the face defined by vertices (i0,j0), (i0+n,j0), (i0,j0+n)
  // (the three edges are split at their midpoints, and the face is replaced
  // by four faces, which are then themselves subdivided)
  private void subdivideFace(int i0, int j0, int n) {

    assert( n != 0 );
    if ( n == -1 || n == +1 ) return;
    
    interpolateVertices(i0,   j0, i0+n, j0);
    interpolateVertices(i0,   j0, i0,   j0+n);
    interpolateVertices(i0+n, j0, i0,   j0+n);
 
    subdivideFace(i0,     j0,     n/2);
    subdivideFace(i0+n/2, j0,     n/2);
    subdivideFace(i0,     j0+n/2, n/2);
    subdivideFace(i0+n/2, j0+n/2, -n/2);
    
  } // subdivideFace()

  // calculate new vertex coordinates on the edge between (i0,j0) and (i1,j1)
  // (the midpoint of the edge is projected onto the unit sphere)
  private void interpolateVertices(int i0, int j0, int i1, int j1) {

    assert( (i0+i1) % 2 == 0 );
    assert( (j0+j1) % 2 == 0 );
    
    setVertexArray(i0, j0);
    final float x0 = mVertexArray.get(),
                y0 = mVertexArray.get(),
                z0 = mVertexArray.get();
    
    setVertexArray(i1, j1);
    final float x1 = mVertexArray.get(),
                y1 = mVertexArray.get(),
                z1 = mVertexArray.get();

    final float x = (x0 + x1)/2,
                y = (y0 + y1)/2,
                z = (z0 + z1)/2;
    final float d = (float)Math.sqrt(x*x + y*y + z*z);
    assert( d > 1.0e-3f && d <= 1.0f );
    
    setVertexArray((i0+i1)/2, (j0+j1)/2);
    mVertexArray.put(x/d);
    mVertexArray.put(y/d);
    mVertexArray.put(z/d);
    
  } // interpolateVertices()
  
  // set the current position of the vertex array buffer to the vertex (i,j)
  private void setVertexArray(int i, int j) {
    
    mVertexArray.position( 3*vertexIndex(i, j) );
    
  } // setVertexArray()
  
  // return the flattened (one-dimensional) index for the the vertex (i,j)
  private int vertexIndex(int i, int j) {
    
    assert( i >= 0 && i < mNumIVerts );
    assert( j >= 0 && j < mNumIVerts );
    assert( i + j < mNumIVerts );
    
    final int index = i + ((2*mNumIVerts-(j-1))*j)/2;
    assert( index >= 0 && index < mNumVertices );

    return index;
    
  } // vertexIndex()

  // create an array of triangular faces at the specified depth 
  private ShortBuffer makeFaces(int depth) {
    
    assert( depth >= 0 && depth <= mSubdivisionDepth );
    
    final int numFaces = (1 << (2*depth));
    ShortBuffer faceArray = Env.allocateShortBuffer(3*numFaces);
    
    final int delta = (1 << (mSubdivisionDepth - depth));
    for ( int j = 0 ; j < mNumIVerts-delta ; j += delta ) {
      for ( int i = 0 ; i < mNumIVerts-delta-j ; i += delta ) {

        faceArray.put( (short)vertexIndex(i, j) );
        faceArray.put( (short)vertexIndex(i+delta, j) );
        faceArray.put( (short)vertexIndex(i, j+delta) );
        
        if ( i < mNumIVerts-delta-j-1 ) {
          faceArray.put( (short)vertexIndex(i+delta, j) );
          faceArray.put( (short)vertexIndex(i+delta, j+delta) );
          faceArray.put( (short)vertexIndex(i, j+delta) );
        }
        
      }      
    }
    assert( faceArray.position() == 3*numFaces );
    
    return faceArray;
    
  } // makeFaces()
  
  // record texture coordinates for the vertices
  private void makeTexCoords() {
    
    mTexCoordArray = Env.allocateFloatBuffer(2*mNumVertices);
    
    for ( int j = 0 ; j < mNumIVerts ; j++ ) {
      for ( int i = 0 ; i < mNumIVerts-j ; i++ ) {
        final float u = i/(float)(mNumIVerts-1),
                    v = j/(float)(mNumIVerts-1);
        assert( u >= 0 && u <= 1 && v >= 0 && v <= 1 && u+v <= 1 );
        mTexCoordArray.put(u);
        mTexCoordArray.put(v);
      }
    }

    assert( mTexCoordArray.position() == 2*mNumVertices );
    
  } // makeTexCoords()

  // convert texture coordinates (u,v) to three-space coordinates (x,y,z) 
  public void texCoordsToXyz(float u, float v, float xyz[]) {
    
    assert( u >= 0.0f && u <= 1.0f );
    assert( v >= 0.0f && v <= 1.0f );
    assert( u+v <= 1.0f );
    
    assert( xyz != null && xyz.length == 3 );

    final float ix = u*(mNumIVerts-1),
                jx = v*(mNumIVerts-1);
    
    int i0 = (int)Math.floor(ix),
        j0 = (int)Math.floor(jx);
    float di = ix - i0,
          dj = jx - j0;
    
    if ( i0 == mNumIVerts-1 ) {
      assert( j0 == 0 && dj < 1.0e-4f );
      i0 -= 1;
      di = 1.0f;
      dj = 0.0f;
    } else if ( j0 == mNumIVerts-1 ) {
      assert( i0 == 0 && di < 1.0e-4f );
      j0 -= 1;
      di = 0.0f;
      dj = 1.0f;
    } else if ( i0+j0 == mNumIVerts-1 ) {
      assert( di < 1.0e-4f && dj < 1.0e-4f );
      i0 -= 1;
      j0 -= 1;
      di = dj = 1.0f;
    }

    setVertexArray(i0+1, j0);
    final float xi = mVertexArray.get(),
                yi = mVertexArray.get(),
                zi = mVertexArray.get();

    setVertexArray(i0, j0+1);
    final float xj = mVertexArray.get(),
                yj = mVertexArray.get(),
                zj = mVertexArray.get();
    
    if ( di+dj > 1.0f && i0+j0+2 <= mNumIVerts-1 ) {
      i0 += 1;
      j0 += 1;
      final float ddi = di;
      di = 1.0f - dj;
      dj = 1.0f - ddi;
    }
    
    setVertexArray(i0, j0);
    final float x0 = mVertexArray.get(),
                y0 = mVertexArray.get(),
                z0 = mVertexArray.get();
    
    // note: these coordinates could be normalized to lie on the unit sphere 
    xyz[0] = x0 + di*(xi - x0) + dj*(xj - x0);
    xyz[1] = y0 + di*(yi - y0) + dj*(yj - y0);
    xyz[2] = z0 + di*(zi - z0) + dj*(zj - z0);
    
  } // texCoordsToXyz()

  // ensure that buffer objects are initialized
  public void prepareBufferObjs(GL2 gl, int depth) {
    
    assert( depth >= 0 && depth <= mSubdivisionDepth );
    
    if ( mFaceBufferObj[depth] != 0 ) return;
    
    final int numBuffers = ( (mVertexBufferObj == 0) ? 3 : 1 );
    int bufferIds[] = new int[numBuffers];
    gl.glGenBuffers(numBuffers, bufferIds, 0);
    
    if ( mVertexBufferObj == 0 ) {
      assert( numBuffers == 3 );
      mVertexBufferObj = bufferIds[1];
      gl.glBindBuffer(GL.GL_ARRAY_BUFFER, mVertexBufferObj);
      mVertexArray.rewind();
      gl.glBufferData(GL.GL_ARRAY_BUFFER, 4*mVertexArray.limit(), 
                      mVertexArray, GL.GL_STATIC_DRAW);
    }
    
    if ( mTexCoordBufferObj == 0 ) {
      assert( numBuffers == 3 );
      mTexCoordBufferObj = bufferIds[2];
      gl.glBindBuffer(GL.GL_ARRAY_BUFFER, mTexCoordBufferObj);
      mTexCoordArray.rewind();
      gl.glBufferData(GL.GL_ARRAY_BUFFER, 4*mTexCoordArray.limit(), 
                      mTexCoordArray, GL.GL_STATIC_DRAW);
    }

    mFaceBufferObj[depth] = bufferIds[0];
    gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, mFaceBufferObj[depth]);
    ShortBuffer faceArray = mFaceArrays[depth];
    faceArray.rewind();
    gl.glBufferData(GL.GL_ELEMENT_ARRAY_BUFFER, 2*faceArray.limit(), 
                    faceArray, GL.GL_STATIC_DRAW);    
    
  } // prepareBufferObjs()
  
  // display the octant at the specified subdivision depth
  public void display(GL2 gl, int depth) {

    assert( depth >= 0 && depth <= mSubdivisionDepth );
    if ( mFaceArrays[depth] == null ) mFaceArrays[depth] = makeFaces(depth);
    
    final int method = 1;
    switch (method) {
      case 0: {
        displayDirect(gl, depth);
      } break;
      case 1: {
        displayArrays(gl, depth);
      } break;
      case 2: {
        displayBufferObjs(gl, depth);
      } break;
      case 3: {
        bindData(gl,depth);
        displayData(gl);
        unbindData(gl);
      } break;
    }
    
  } // display()
  
  // display the octant directly as a list of triangles
  private void displayDirect(GL2 gl, int depth) {

    assert( depth >= 0 && depth <= mSubdivisionDepth );
    
    ShortBuffer faceArray = mFaceArrays[depth];
    faceArray.rewind();
    
    gl.glBegin(GL.GL_TRIANGLES);
    for ( int k = 0 ; k < faceArray.limit() ; k++ ) {
      final int vIndex = faceArray.get();
      mVertexArray.position(3*vIndex);
      final float x = mVertexArray.get(),
                  y = mVertexArray.get(),
                  z = mVertexArray.get();
      mTexCoordArray.position(2*vIndex);
      final float u = mTexCoordArray.get(),
                  v = mTexCoordArray.get();
      gl.glTexCoord2f(u, v);
      gl.glNormal3f(x, y, z);
      gl.glVertex3f(x, y, z);
    }
    gl.glEnd();
    
  } // displayDirect()
  
  // display the octant using vertex and face arrays
  private void displayArrays(GL2 gl, int depth) {

    assert( depth >= 0 && depth <= mSubdivisionDepth );

    mVertexArray.rewind();
    mTexCoordArray.rewind();
    
    gl.glEnableClientState(GL2.GL_VERTEX_ARRAY);
    gl.glEnableClientState(GL2.GL_NORMAL_ARRAY);
    gl.glEnableClientState(GL2.GL_TEXTURE_COORD_ARRAY);

    gl.glVertexPointer(3, GL.GL_FLOAT, 0, mVertexArray);
    gl.glNormalPointer(GL.GL_FLOAT, 0, mVertexArray);
    gl.glTexCoordPointer(2, GL.GL_FLOAT, 0, mTexCoordArray);
    
    ShortBuffer faceArray = mFaceArrays[depth];
    faceArray.rewind();

    gl.glDrawElements(GL.GL_TRIANGLES, faceArray.limit(), 
                      GL.GL_UNSIGNED_SHORT, faceArray);
    
    gl.glDisableClientState(GL2.GL_TEXTURE_COORD_ARRAY);
    gl.glDisableClientState(GL2.GL_NORMAL_ARRAY);
    gl.glDisableClientState(GL2.GL_VERTEX_ARRAY);
    
  } // displayArrays()
  
  // display the octant using buffer objects
  private void displayBufferObjs(GL2 gl, int depth) {

    assert( depth >= 0 && depth <= mSubdivisionDepth );

    prepareBufferObjs(gl, depth);
    
    gl.glEnableClientState(GL2.GL_VERTEX_ARRAY);
    gl.glEnableClientState(GL2.GL_NORMAL_ARRAY);
    gl.glEnableClientState(GL2.GL_TEXTURE_COORD_ARRAY);

    gl.glBindBuffer(GL.GL_ARRAY_BUFFER, mVertexBufferObj);
    gl.glVertexPointer(3, GL.GL_FLOAT, 0, 0);
    gl.glNormalPointer(GL.GL_FLOAT, 0, 0);
    
    gl.glBindBuffer(GL.GL_ARRAY_BUFFER, mTexCoordBufferObj);
    gl.glTexCoordPointer(2, GL.GL_FLOAT, 0, 0);
    
    final int numFaces = ( 1 << (2*depth) );
    gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, mFaceBufferObj[depth]);
    gl.glDrawElements(GL.GL_TRIANGLES, 3*numFaces, 
                      GL.GL_UNSIGNED_SHORT, 0);
  
    gl.glDisableClientState(GL2.GL_TEXTURE_COORD_ARRAY);
    gl.glDisableClientState(GL2.GL_NORMAL_ARRAY);
    gl.glDisableClientState(GL2.GL_VERTEX_ARRAY);
    
  } // displayBufferObjs()
  
  // bind the buffer objects prior to calling displayData()
  public void bindData(GL2 gl, int depth) {
    
    assert( depth >= 0 && depth <= mSubdivisionDepth );
    if ( mFaceArrays[depth] == null ) mFaceArrays[depth] = makeFaces(depth);

    prepareBufferObjs(gl, depth);

    assert( mBoundDepth == -1 );
    mBoundDepth = depth;

    gl.glEnableClientState(GL2.GL_VERTEX_ARRAY);
    gl.glEnableClientState(GL2.GL_NORMAL_ARRAY);
    gl.glEnableClientState(GL2.GL_TEXTURE_COORD_ARRAY);

    gl.glBindBuffer(GL.GL_ARRAY_BUFFER, mVertexBufferObj);
    gl.glVertexPointer(3, GL.GL_FLOAT, 0, 0);
    gl.glNormalPointer(GL.GL_FLOAT, 0, 0);
    
    gl.glBindBuffer(GL.GL_ARRAY_BUFFER, mTexCoordBufferObj);
    gl.glTexCoordPointer(2, GL.GL_FLOAT, 0, 0);
    
    ShortBuffer faceArray = mFaceArrays[depth];
    faceArray.rewind();
    
    gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, mFaceBufferObj[depth]);
        
  } // bindData()

  // display the buffer objects prepared by bindData()
  public void displayData(GL2 gl) {

    assert( mBoundDepth >= 0 );
    final int numFaces = (1 << (2*mBoundDepth));
    gl.glDrawElements(GL.GL_TRIANGLES, 3*numFaces, GL.GL_UNSIGNED_SHORT, 0);
    
  } // displayData()
  
  // clean-up after calling bindData() and displayData()
  public void unbindData(GL2 gl) {
    
    assert( mBoundDepth >= 0 );
    mBoundDepth = -1;
    
    gl.glDisableClientState(GL2.GL_TEXTURE_COORD_ARRAY);
    gl.glDisableClientState(GL2.GL_NORMAL_ARRAY);
    gl.glDisableClientState(GL2.GL_VERTEX_ARRAY);
    
  } // unbindData()
  
} // class Octant
