/**
 *  Sphere.java
 *  Copyright Simon Hern 2009
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package Space.planets;

import Space.Env;

import javax.media.opengl.*;
import com.sun.opengl.util.texture.*;

// Approximation of the unit sphere based on subdivision of an octahedron.
// 
// The Octant class handles the subdivision of the shape, with eight Octants
// being displayed to produce a full sphere.
//
// The Sphere may be textured, with a single square texture being divided
// into eight parts corresponding to the eight Octants.  The assignment of
// texture coordinates is a little fiddly.  The texture is first divided
// into four identical square parts corresponding to the quadrants of the
// sphere {x>=0,z>=0}, {x>=0,z<0}, {x<0,z<0}, {x<0,z>=0}.  Each of these
// quadrants is represented using two Octants, a north one (y>=0) and a
// south one (y<0).  The square texture for a quadrant is split down its
// diagonal to give two triangular bits of texture for the two Octants.
//
// The textures for the two Octants in a quadrant join up exactly along the
// diagonal (corresponding to the Sphere's equator).  However, the textures
// for the quadrants are separate, with coordinates that end at texel 
// centres (rather than edges).  This is done so that the four quadrants
// can be joined up precisely, although it does mean that some lines of
// texels in the texture are unnecessarily repeated.

public class Sphere {
  
  // a single, static object should do for most purposes
  static private final int kSingletonMinDepth = 0,
                           kSingletonMaxDepth = 6;
  static private Sphere kSingleton = new Sphere(kSingletonMinDepth,
                                                kSingletonMaxDepth);
  static public Sphere getInstance() { return kSingleton; }
  
  //---
  
  // object representing an eighth of a sphere
  private Octant mOctant;
  
  // constructor
  public Sphere(int minDepth, int maxDepth) {
    
    assert( minDepth >= 0 );
    assert( maxDepth >= minDepth );
    mOctant = new Octant(minDepth, maxDepth);

  } // constructor

  // the maximum subdivision depth for this shape
  public int maxDepth() { return mOctant.maxDepth(); }
  
  // map texel (i,j) to three-space coordinates (x,y,z) 
  // (the texture is assumed to be a square with the specified width) 
  public void texelToXyz(int i, int j, int texWidth, float xyz[]) {
    
    assert( texWidth > 0 && texWidth%2 == 0 );
    
    assert( i >= 0 && i < texWidth );
    assert( j >= 0 && j < texWidth );
    
    assert( xyz != null && xyz.length == 3 );

    int quadrant = ( (i < texWidth/2) ? 0 : 1 )
                 + ( (j < texWidth/2) ? 0 : 2 );
    i %= (texWidth/2);
    j %= (texWidth/2);
    
    boolean southHemi = false;
    if ( i+j > texWidth/2-1 ) {
      int ii = i;
      i = (texWidth/2-1) - j;
      j = (texWidth/2-1) - ii;
      southHemi = true;
    }
    
    assert( i >= 0 && i < texWidth/2 );
    assert( j >= 0 && j < texWidth/2 );
    assert( i+j < texWidth/2 );

    float u = i/(float)(texWidth/2 - 1),
          v = j/(float)(texWidth/2 - 1);
    mOctant.texCoordsToXyz(u, v, xyz);
    
    if ( quadrant == 1 ) {
      final float x = xyz[0], 
                  z = xyz[2];
      xyz[0] = +z;
      xyz[2] = -x;
    } else if ( quadrant == 2 ) {
      xyz[0] = -xyz[0];
      xyz[2] = -xyz[2];
    } else if ( quadrant == 3 ) {
      final float x = xyz[0], 
                  z = xyz[2];
      xyz[0] = -z;
      xyz[2] = +x;      
    }
    if ( southHemi ) xyz[1] = -xyz[1];
    
  } // texelToXyz()  
  
  // display the sphere at the specified depth of subdivision
  public void display(GL2 gl, int depth, Texture texture) {
    
    assert( depth >= 0 && depth <= maxDepth() );

    if ( Env.useBufferObjs() ) mOctant.bindData(gl, depth);
    
    gl.glShadeModel(GL2.GL_SMOOTH);

    float diffuse[] = { 1.0f, 1.0f, 1.0f, 1.0f };
    float specular[] = { 0.0f, 0.0f, 0.0f, 1.0f };
    gl.glMaterialfv(GL.GL_FRONT, GL2.GL_AMBIENT_AND_DIFFUSE, diffuse, 0);
    gl.glMaterialfv(GL.GL_FRONT, GL2.GL_SPECULAR, specular, 0);

    int   texWidth  = 0;
    float texScale  = 1.0f,
          texOffset = 0.0f;
    if ( texture != null ) {
      texWidth = texture.getWidth();
      texScale = (texWidth/2-1)/(float)texWidth;
      texOffset = 0.5f/texWidth;
      
      texture.enable();
      texture.bind();
      gl.glTexEnvf(GL2.GL_TEXTURE_ENV, GL2.GL_TEXTURE_ENV_MODE, 
                   GL2.GL_MODULATE);
    }
    
    for ( int octant = 0 ; octant < 8 ; octant++ ) {
      
      final int     quadrant  = octant/2;
      final boolean southHemi = (octant%2 == 1);
      
      if ( texture != null ) {
        gl.glMatrixMode(GL.GL_TEXTURE);
        gl.glPushMatrix();
        gl.glLoadIdentity();
        
        float uShift = ( quadrant%2 == 1 ) ? 0.5f : 0.0f,
              vShift = ( quadrant/2 == 1 ) ? 0.5f : 0.0f;
        gl.glTranslatef(texOffset+uShift, texOffset+vShift, 0.0f);
        gl.glScalef(texScale, texScale, 1.0f);
        
        if ( southHemi ) {
          gl.glTranslatef(1.0f, 1.0f, 0.0f);
          gl.glScalef(-1.0f, -1.0f, 1.0f);
        }
        
        gl.glMatrixMode(GL2.GL_MODELVIEW);
      }

      gl.glPushMatrix();
      gl.glRotatef(90.0f*(octant/2), 0.0f, 1.0f, 0.0f);
      if ( southHemi ) {
        gl.glRotatef(90.0f, 0.0f, 1.0f, 0.0f);
        gl.glRotatef(180.0f, 0.0f, 0.0f, 1.0f);
      }
      if ( Env.useBufferObjs() ) mOctant.displayData(gl);
      else                       mOctant.display(gl, depth);
      gl.glPopMatrix();

      if ( texture != null ) {
        gl.glMatrixMode(GL.GL_TEXTURE);
        gl.glPopMatrix();
        gl.glMatrixMode(GL2.GL_MODELVIEW);
      }
        
    } // for (octant)

    if ( texture != null ) {
      texture.disable();
    }
    if ( Env.useBufferObjs() ) mOctant.unbindData(gl);

  } // display()

} // class Sphere
