/**
 *  MainWindow.java
 *  Copyright Simon Hern 2009
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package Space.planets.viewer;

import java.awt.*;
import java.awt.event.*;
import javax.media.opengl.*;
import javax.media.opengl.awt.*;

import com.sun.opengl.util.*;

// the frame for the planet viewer
public class MainWindow extends Frame implements GLEventListener {

  // main
  public static void main(String[] args) {
    
    new MainWindow();
    
  } // main()

  // the JOGL animation object
  private Animator mAnimator;
  
  // the planet viewer object
  private Viewer mViewer;
  
  // time of first call to display (nanoseconds)
  private long mStartNanos;
  
  // time of previous call to display (nanoseconds)
  private long mPrevNanos;
  
  // target time at which to update statistics (nanoseconds)
  private long mNextUpdateTime;
  
  // previous time at which statistics were updated (nanoseconds)
  private long mPrevUpdateTime;
  
  // total number of calls to display
  private int mNumFramesTotal;
  
  // number of calls to display since the last update
  private int mNumFramesSinceUpdate;
  
  // constructor (sets up the viewer window)
  public MainWindow() {
  
    mStartNanos = mPrevNanos = mNextUpdateTime = mPrevUpdateTime = -1;
    mNumFramesTotal = mNumFramesSinceUpdate = 0;
    
    GLCapabilities caps = new GLCapabilities(GLProfile.get(GLProfile.GL2));
    GLCanvas canvas = new GLCanvas(caps);
    canvas.addGLEventListener(this);
    add(canvas);
    
    mAnimator = new Animator(canvas); //new FPSAnimator(canvas, 60);
    mAnimator.setRunAsFastAsPossible(true);
    
    int nx = 600, ny = 600;
    setSize(nx, ny);

    Dimension screenDim = Toolkit.getDefaultToolkit().getScreenSize();
    setLocation(screenDim.width/2 - nx/2, screenDim.height/2 - ny/2);
    
    setTitle("Planet Viewer");
    
    addWindowListener(
      new WindowAdapter() {
        public void windowClosing(WindowEvent e) { exit(); }
        public void windowDeiconified(WindowEvent e) { start(); }
        public void windowIconified(WindowEvent e) { stop(); }
      }
    );
    
    addMouseListener(
      new MouseAdapter() {
        public void mouseClicked(MouseEvent e) { requestFocusInWindow(); }
      }
    );
    canvas.addMouseListener(
      new MouseAdapter() {
        public void mouseClicked(MouseEvent e) { requestFocusInWindow(); }
      }
    );
    requestFocusInWindow();

    mViewer = new Viewer();

    setVisible(true);
    
    mAnimator.start();

  } // constructor
  
  // start the application
  public void start() {

    requestFocusInWindow();

  } // start()
  
  // interrupt the application
  public void stop() {
    
    if ( mNumFramesTotal > 0 ) {
      float secondsTotal = (System.nanoTime() - mStartNanos)*1.0e-9f;
      float framesPerSecond = mNumFramesTotal / secondsTotal;
      System.out.println("Overall: " 
                         + String.format("%.1f", framesPerSecond)
                         + " fps (" + mNumFramesTotal + " frames in "
                         + String.format("%.3f", secondsTotal)
                         + " seconds)");
    }
    
  } // stop()

  // end the application
  public void exit() {
 
    stop();
    // make sure the call to Animator.stop() completes before exiting
    new Thread(
      new Runnable() {
        public void run() { 
          mAnimator.stop(); 
          System.exit(0); }
      }
    ).start();
    
  } // exit()
  
  // GL initialize
  @Override
  public void init(GLAutoDrawable drawable) {
    
    drawable.setGL(new DebugGL2(drawable.getGL().getGL2()));
    //drawable.setGL(new TraceGL(drawable.getGL(), System.out));

    GL2 gl = drawable.getGL().getGL2();

    System.out.println("Initializing JOGL");
    System.out.println("GL class: " + gl.getClass().getName());
    System.out.println("GLCapabilities: " 
                       + drawable.getChosenGLCapabilities());
    System.out.println("GL_VENDOR: " + gl.glGetString(GL.GL_VENDOR));
    System.out.println("GL_RENDERER: " + gl.glGetString(GL.GL_RENDERER));
    System.out.println("GL_VERSION: " + gl.glGetString(GL.GL_VERSION));

    gl.setSwapInterval(1);

    mViewer.init(gl);
    
  } // GLEventListener.init()

  // GL dispose
  @Override
  public void dispose(GLAutoDrawable drawable) {
    
  } // GLEventListener.dispose()
  
  // GL change of window size
  @Override
  public void reshape(GLAutoDrawable drawable, 
                      int x, int y, int width, int height) {
    
    width = Math.max(width, 1);
    height = Math.max(height, 1);
    
    GL2 gl = drawable.getGL().getGL2();
    gl.glViewport(0, 0, width, height);
    mViewer.reshape(gl, width, height);
    
  } // GLEventListener.reshape()

  // main GL display routine
  @Override
  public void display(GLAutoDrawable drawable) {
  
    GL2 gl = drawable.getGL().getGL2();

    long newNanos = System.nanoTime();
    boolean firstTime = ( mStartNanos == -1 );

    // advance
    if ( firstTime ) {
      mViewer.advance(0.0f);
    } else {
      float dt = (newNanos-mPrevNanos)*1.0e-9f;
      assert( dt >= 0.0f );
      mViewer.advance(dt);
    }
    
    // display
    mViewer.display(gl);
    gl.glFlush();

    // update statistics
    if ( firstTime ) {
      mStartNanos = newNanos;
      mPrevUpdateTime = newNanos;
      mNextUpdateTime = newNanos + 1000000000L;
      mNumFramesTotal = mNumFramesSinceUpdate = 0;
    } else {
      mNumFramesTotal += 1;
      mNumFramesSinceUpdate += 1;
      if ( newNanos > mNextUpdateTime ) {
        float secondsSinceUpdate = (newNanos - mPrevUpdateTime)*1.0e-9f;
        float framesPerSecond = mNumFramesSinceUpdate / secondsSinceUpdate;
        System.out.println(String.format("%.1f", framesPerSecond)
                           + " fps (" + mNumFramesSinceUpdate + " frames in "
                           + String.format("%.3f", secondsSinceUpdate)
                           + " seconds)");
        mPrevUpdateTime = newNanos;
        mNextUpdateTime = newNanos + 1000000000L;
        mNumFramesSinceUpdate = 0;
      }
    }
    mPrevNanos = newNanos;

  } // GLEventListener.display()

} // class MainWindow
