/**
 *  Viewer.java
 *  Copyright Simon Hern 2009
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package Space.planets.viewer;

import Space.planets.*;
import javax.media.opengl.*;

// render a spinning planet for test purposes
public class Viewer {

  // light parameters
  static private final float kLightAmbient[]    = {0.0f, 0.0f, 0.0f, 1.0f};
  static private final float kLight0Diffuse[]   = {0.8f, 0.8f, 0.8f, 1.0f}, 
                             kLight0Ambient[]   = {0.2f, 0.2f, 0.2f, 1.0f}, 
                             kLight0Specular[]  = {0.0f, 0.0f, 0.0f, 1.0f}, 
                             kLight0Direction[] = {-0.5f, +0.6f, +0.5f, 0.0f};

  // planet is titled at an angle (degrees)
  static private final float kTiltAngle = +60.0f;

  // speed of planet rotation (degrees per second)
  static private final float kAngSpeed = 20.0f;
  
  // height of viewport (pixels)
  private int mWindowHeight;
  
  // object representing the sphere
  private Globe mGlobe;
  
  // size of planet relative to height of display
  private float mScale;
  
  // current angle of rotation (degrees)
  private float mAngle;

  // constructor
  public Viewer() {
    
    //mGlobe = new Globe(new FieryOrangeSkin(), 256);
    //mGlobe = new Globe(new CloudySkin(), 256);
    //mGlobe = new Globe(new CellularSkin(), 256);
    mGlobe = new Globe(new RoughGreySkin(), 256);

    mScale = 0.95f;
    mAngle = 0.0f;
  
    mWindowHeight = 1;
    
  } // constructor

  // some preparations
  public void init(GL2 gl) {
    
    gl.glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

    gl.glEnable(GL.GL_CULL_FACE);
      
    gl.glEnable(GL.GL_DEPTH_TEST);
    gl.glEnable(GL2.GL_RESCALE_NORMAL);

    gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);

    addLights(gl);

    mGlobe.prepare(gl);
    
  } // init()

  // specify lighting details
  private void addLights(GL2 gl) {
    
    gl.glEnable(GL2.GL_LIGHTING);

    gl.glLightModelfv(GL2.GL_LIGHT_MODEL_AMBIENT, kLightAmbient, 0);
    
    gl.glEnable(GL2.GL_LIGHT0);
    gl.glLightfv(GL2.GL_LIGHT0, GL2.GL_DIFFUSE, kLight0Diffuse, 0);
    gl.glLightfv(GL2.GL_LIGHT0, GL2.GL_AMBIENT, kLight0Ambient, 0);
    gl.glLightfv(GL2.GL_LIGHT0, GL2.GL_SPECULAR, kLight0Specular, 0);
    gl.glLightfv(GL2.GL_LIGHT0, GL2.GL_POSITION, kLight0Direction, 0);
    
  } // addLights()

  // redefine viewport when window change shape (orthographic projection)
  public void reshape(GL2 gl, int width, int height) {
    
    final float h = width/(float)height;

    gl.glViewport(0, 0, width, height);
    gl.glMatrixMode(GL2.GL_PROJECTION);
    gl.glLoadIdentity();
    gl.glOrtho(-h, +h, -1.0f, +1.0f, -10.0f, +10.0f);
    gl.glMatrixMode(GL2.GL_MODELVIEW);
    gl.glLoadIdentity();
      
    mWindowHeight = height;
    
  } // reshape()

  // update rotation, etc.
  public void advance(float dt) {
    
    mAngle += kAngSpeed*dt;

    final float radiusPixels = 0.5f*mScale*mWindowHeight;
    mGlobe.setDepthFromSize(radiusPixels);
    
  } // advance()

  // display the planet
  public void display(GL2 gl) {

    final float viewXyz[] = { 0.0f, 0.0f, 1.0f };
    
    gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
    gl.glLoadIdentity();
    
    gl.glPushMatrix();
      gl.glScalef(mScale, mScale, mScale);
      mGlobe.displayHalo(gl, 1.1f, viewXyz);
      gl.glRotatef(kTiltAngle, 1.0f, 0.0f, 0.0f);
      gl.glRotatef(mAngle, 0.0f, 1.0f, 0.0f);
      mGlobe.display(gl);
    gl.glPopMatrix();

  } // display()
  
} // class Viewer
